# -*- coding: utf-8 -*-
"""
Expressions to calculate Debye-Waller factor.

For details see documentation.
"""

import numpy

from .unit_cell import calc_reciprocal_by_unit_cell_parameters

na = numpy.newaxis

def calc_beta_by_u_ij(
        u_ij, reciprocal_unit_cell_parameters,
        flag_u_ij: bool = False):
    r"""
    Calculate beta.

    u_11, u_22, u_33, u_12, u_13, u_23 = u_i

    cell is Cell class

    Output:
        $\beta_{11}$, $\beta_{22}$, $\beta_{33}$
        $\beta_{12}$, $\beta_{13}$, $\beta_{23}$
    """
    u_11, u_22, u_33 = u_ij[0], u_ij[1], u_ij[2]
    u_12, u_13, u_23 = u_ij[3], u_ij[4], u_ij[5]

    ia = reciprocal_unit_cell_parameters[0]
    ib = reciprocal_unit_cell_parameters[1]
    ic = reciprocal_unit_cell_parameters[2]
    beta_11 = 2.*numpy.pi**2*u_11*ia**2
    beta_22 = 2.*numpy.pi**2*u_22*ib**2
    beta_33 = 2.*numpy.pi**2*u_33*ic**2
    beta_12 = 2.*numpy.pi**2*u_12*ia*ib
    beta_13 = 2.*numpy.pi**2*u_13*ia*ic
    beta_23 = 2.*numpy.pi**2*u_23*ib*ic
    beta = numpy.stack([beta_11, beta_22, beta_33, beta_12, beta_13, beta_23], axis=0)
    dder = {}
    if flag_u_ij:
        dder["u_ij"] = numpy.stack([
            2.*numpy.pi**2*numpy.ones_like(u_11)*ia**2,
            2.*numpy.pi**2*numpy.ones_like(u_22)*ib**2,
            2.*numpy.pi**2*numpy.ones_like(u_33)*ic**2,
            2.*numpy.pi**2*numpy.ones_like(u_12)*ia*ib,
            2.*numpy.pi**2*numpy.ones_like(u_13)*ia*ic,
            2.*numpy.pi**2*numpy.ones_like(u_23)*ib*ic], axis=0)
    return beta, dder



def calc_u_ij_by_beta(
        beta, reciprocal_unit_cell_parameters,
        flag_beta: bool = False):
    r"""
    Calculate beta.

    beta_11, beta_22, beta_33, beta_12, beta_13, beta_23 = beta_i

    cell is Cell class

    Output:
        $\u_{11}$, $\u_{22}$, $\u_{33}$
        $\u_{12}$, $\u_{13}$, $\u_{23}$
    """
    beta_11, beta_22, beta_33 = beta[0], beta[1], beta[2]
    beta_12, beta_13, beta_23 = beta[3], beta[4], beta[5]
    pi_sq = numpy.pi**2
    ia = reciprocal_unit_cell_parameters[0]
    ib = reciprocal_unit_cell_parameters[1]
    ic = reciprocal_unit_cell_parameters[2]
    u_11 = beta_11/(2. * pi_sq * ia**2)
    u_22 = beta_22/(2. * pi_sq * ib**2)
    u_33 = beta_33/(2. * pi_sq * ic**2)
    u_12 = beta_12/(2. * pi_sq * ia*ib)
    u_13 = beta_13/(2. * pi_sq * ia*ic)
    u_23 = beta_23/(2. * pi_sq * ib*ic)
    u_ij = numpy.stack([u_11, u_22, u_33, u_12, u_13, u_23], axis=0)
    dder = {}
    if flag_beta:
        dder["beta"] = numpy.stack([
            numpy.ones_like(beta_11)/(2. * pi_sq * ia**2),
            numpy.ones_like(beta_22)/(2. * pi_sq * ib**2),
            numpy.ones_like(beta_33)/(2. * pi_sq * ic**2),
            numpy.ones_like(beta_12)/(2. * pi_sq * ia*ib),
            numpy.ones_like(beta_13)/(2. * pi_sq * ia*ic),
            numpy.ones_like(beta_23)/(2. * pi_sq * ib*ic)], axis=0)
    return u_ij, dder


def calc_b_iso_beta(
        unit_cell_parameters, atom_adp_type, atom_iso_param, atom_aniso_param):
    """
    Calculate b_iso and beta_ij based on atom_site and atom_sites.
    For each atom defined in atom_site.
    """
    coeff = float(8.*numpy.square(numpy.pi))

    reciprocal_unit_cell_parameters, dder = calc_reciprocal_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)

    np_atom_adp_type, np_atom_iso_param, np_atom_aniso_param = \
        numpy.atleast_1d(atom_adp_type, atom_iso_param, atom_aniso_param)
    
    f_uiso = np_atom_adp_type == "Uiso"
    f_biso = np_atom_adp_type == "Biso"
    f_uovl = np_atom_adp_type == "Uovl"
    f_umpe = np_atom_adp_type == "Umpe"
    f_uani = np_atom_adp_type == "Uani"
    f_bovl = np_atom_adp_type == "Bovl"
    f_bani = np_atom_adp_type == "Bani"
    b_iso = numpy.zeros_like(np_atom_iso_param)
    beta = numpy.zeros_like(np_atom_aniso_param)

    b_iso[f_uiso] = coeff*np_atom_iso_param[f_uiso]
    b_iso[f_biso] = np_atom_iso_param[f_biso]
    b_iso[f_uovl] = coeff*np_atom_iso_param[f_uovl] #FIXME: redo it
    b_iso[f_umpe] = coeff*np_atom_iso_param[f_umpe] #FIXME: redo it
    b_iso[f_bovl] = np_atom_iso_param[f_bovl] #FIXME: redo it

    beta[:, f_uani], dder_beta = calc_beta_by_u_ij(
        np_atom_aniso_param[:, f_uani], reciprocal_unit_cell_parameters, flag_u_ij=False)

    beta[:, f_bani] = np_atom_aniso_param[:, f_bani]
    return b_iso, beta


def calc_param_iso_aniso_by_b_iso_beta(unit_cell_parameters, atom_adp_type, b_iso, beta):
    coeff = 1./float(8.*numpy.square(numpy.pi))

    reciprocal_unit_cell_parameters, dder = calc_reciprocal_by_unit_cell_parameters(
        unit_cell_parameters, flag_unit_cell_parameters=False)

    np_atom_adp_type, np_b_iso, np_beta = \
        numpy.atleast_1d(atom_adp_type, b_iso, beta)

    f_uiso = np_atom_adp_type == "Uiso"
    f_biso = np_atom_adp_type == "Biso"
    f_uovl = np_atom_adp_type == "Uovl"
    f_umpe = np_atom_adp_type == "Umpe"
    f_uani = np_atom_adp_type == "Uani"
    f_bovl = np_atom_adp_type == "Bovl"
    f_bani = np_atom_adp_type == "Bani"

    atom_iso_param = numpy.zeros_like(np_b_iso)
    atom_aniso_param = numpy.zeros_like(np_beta)

    atom_iso_param[f_uiso] = coeff*np_b_iso[f_uiso]
    atom_iso_param[f_biso] = b_iso[f_biso]
    atom_iso_param[f_uovl] = coeff*np_b_iso[f_uovl] #FIXME: redo it
    atom_iso_param[f_umpe] = coeff*np_b_iso[f_umpe] #FIXME: redo it
    atom_iso_param[f_bovl] = np_b_iso[f_bovl] #FIXME: redo it

    atom_aniso_param[:, f_bani] = np_beta[:, f_bani]
    atom_aniso_param[:, f_uani] = calc_u_ij_by_beta(
        np_beta[:, f_uani], reciprocal_unit_cell_parameters, flag_beta=False)[0]

    return atom_iso_param, atom_aniso_param


def calc_power_dwf_iso(sthovl, b_iso, flag_sthovl: bool = False,
                       flag_b_iso: bool = False):
    """Isotropic harmonic Debye-Waller factor.
    """
    sthovl_sq = numpy.square(sthovl)
    power_dwf_iso = b_iso*sthovl_sq
    dder = {}
    if flag_b_iso:
        dder["b_iso"] = sthovl_sq * numpy.ones_like(b_iso)
    if flag_sthovl:
        dder["sthovl"] = 2*b_iso*sthovl*numpy.ones_like(sthovl)
    return power_dwf_iso, dder


def calc_power_dwf_aniso(
        index_hkl, beta, symm_elems_r, flag_beta: bool = False):
    """Anisotropic harmonic Debye-Waller factor.
    """
    b_11, b_22, b_33 = beta[0], beta[1], beta[2]
    b_12, b_13, b_23 = beta[3], beta[4], beta[5]

    h = index_hkl[0]
    k = index_hkl[1]
    l = index_hkl[2]

    r_11, r_12, r_13 = symm_elems_r[0], symm_elems_r[1], symm_elems_r[2]
    r_21, r_22, r_23 = symm_elems_r[3], symm_elems_r[4], symm_elems_r[5]
    r_31, r_32, r_33 = symm_elems_r[6], symm_elems_r[7], symm_elems_r[8]

    h_s = h*r_11 + k*r_21 + l*r_31
    k_s = h*r_12 + k*r_22 + l*r_32
    l_s = h*r_13 + k*r_23 + l*r_33

    power_dwf_aniso = (b_11*numpy.square(h_s) + b_22*numpy.square(k_s) +
                       b_33*numpy.square(l_s) + 2.*b_12*h_s*k_s +
                       2.*b_13*h_s*l_s + 2.*b_23*k_s*l_s)
    dder = {}
    if flag_beta:
        ones_b = numpy.ones_like(b_11)
        der_b_11 = ones_b*numpy.square(h_s)
        der_b_22 = ones_b*numpy.square(k_s)
        der_b_33 = ones_b*numpy.square(l_s)
        der_b_12 = ones_b*2.*h_s*k_s
        der_b_13 = ones_b*2.*h_s*l_s
        der_b_23 = ones_b*2.*k_s*l_s
        dder["beta"] = numpy.stack([der_b_11, der_b_22, der_b_33,
                                    der_b_12, der_b_13, der_b_23], axis=0)
    return power_dwf_aniso, dder


def calc_dwf(index_hkl, sthovl, b_iso, beta, symm_elems_r,
             flag_sthovl: bool = False, flag_b_iso: bool = False,
             flag_beta: bool = False):
    """Calculate Debye-Waller factor."""

    # ["hkl", "atoms"]
    power_iso_2d, dder_power_iso_2d = calc_power_dwf_iso(
        sthovl, b_iso, flag_sthovl=flag_sthovl, flag_b_iso=flag_b_iso)
    #    sthovl[:, na], b_iso[na, :], flag_sthovl=flag_sthovl, flag_b_iso=flag_b_iso)

    # ["hkl", "atoms", "symmetry"]
    power_aniso_3d, dder_power_aniso_3d = calc_power_dwf_aniso(
        index_hkl, beta,
        symm_elems_r, flag_beta=flag_beta)
    #    index_hkl[:, :, na, na], beta[:, na, :, na],
    #    symm_elems_r[:, na, na, :], flag_beta=flag_beta)

    power_3d = power_iso_2d + power_aniso_3d
    #power_3d = power_iso_2d[:, :, na] + power_aniso_3d
    dwf = numpy.exp(-power_3d)
    dder = {}
    if flag_sthovl:
        dder["sthovl"] = None
    if flag_b_iso:
        dder["b_iso"] = None
    if flag_beta:
        dder["beta"] = None
    return dwf, dder




import numpy
from cryspy.A_functions_base.magnetic_form_factor import get_j0_j2_parameters

from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.preocedures import take_items_by_class
from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_1_atom_site import AtomSiteL
from cryspy.C_item_loop_classes.cl_1_atom_site_aniso import AtomSiteAnisoL
from cryspy.C_item_loop_classes.cl_2_atom_site_scat import AtomSiteScatL
from cryspy.C_item_loop_classes.cl_1_atom_type import AtomTypeL
from cryspy.C_item_loop_classes.cl_1_atom_type_scat import AtomTypeScatL
from cryspy.C_item_loop_classes.cl_1_atom_site_susceptibility import AtomSiteSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_atom_electron_configuration import AtomElectronConfigurationL
from cryspy.C_item_loop_classes.cl_2_space_group import SpaceGroup

from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_diffrn_refln import DiffrnReflnL
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import DiffrnRadiation
from cryspy.C_item_loop_classes.cl_2_diffrn_orient_matrix import DiffrnOrientMatrix


from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import ReflnSusceptibilityL


from cryspy.E_data_classes.cl_1_crystal import Crystal
from cryspy.E_data_classes.cl_2_diffrn import Diffrn
from cryspy.E_data_classes.cl_2_pd import Pd



def form_dictionary_by_crystal(data_obj) -> dict:
    """
    Based on cryspy objects form dictionary which contents whole
    information about crystal structure
    """
    if isinstance(data_obj, Crystal):
        ddict = data_obj.get_dictionary()
    else:
        ddict = {}

    return ddict



def form_dictionary_by_diffrn(data_obj) -> dict:
    """
    Based on cryspy objects form dictionary from numpy arrays needed for
    calculations

    Parameters
    ----------
    global_obj : GlobalN of DataN containers
        data container with needed information.

    Returns
    -------
    ddict : dictionary
        output keys: .

    """
    if isinstance(data_obj, Diffrn):
        ddict = data_obj.get_dictionary()
    else:
        ddict = {}

    return ddict



def form_dictionary_by_pd(global_obj) -> dict:
    """
    Based on cryspy objects form dictionary from numpy arrays needed for
    calculations

    Parameters
    ----------
    global_obj : GlobalN of DataN containers
        data container with needed information.

    Returns
    -------
    ddict : dictionary
        output keys: .

    """
    if isinstance(global_obj, Pd):
        global_obj.form_object()

    ddict = {}
    chi2, diffrn_radiation = None, None
    exclude, pd_background = None, None
    pd_instr_reflex_asymmetry, pd_instr_resolution = None, None,
    pd_meas, pd_peak = None, None
    pd_proc, phase = None, None
    range_, refine_ls = None, None
    refln, refln_susceptibility = None, None
    setup = None

    l_obj = take_items_by_class(global_obj, (Setup, ))
    if len(l_obj) > 0:
        setup = l_obj[0]

    l_obj = take_items_by_class(global_obj, (DiffrnRadiation, ))
    if len(l_obj) > 0:
        diffrn_radiation = l_obj[0]

    l_obj = take_items_by_class(global_obj, (ReflnL, ))
    if len(l_obj) > 0:
        refln = l_obj

    l_obj = take_items_by_class(global_obj, (ReflnSusceptibilityL, ))
    if len(l_obj) > 0:
        refln_susceptibility = l_obj


    ddict["name"] = global_obj.get_name()
    if setup is not None:
        ddict["magnetic_field"] = numpy.atleast_1d(setup.field)
        ddict["offset_ttheta"] = numpy.atleast_1d(setup.offset_ttheta)
        ddict["wavelength"] = numpy.atleast_1d(setup.wavelength)

    if diffrn_radiation is not None:
        ddict["beam_polarization"] = numpy.atleast_1d(diffrn_radiation.polarization)
        ddict["flipper_efficiency"] = numpy.atleast_1d(diffrn_radiation.efficiency)

    if refln is not None:
        for refln_phase in refln:
            phase_name = refln_phase.loop_name
            if (refln_phase.is_attribute("index_h") and refln_phase.is_attribute("index_k") and refln_phase.is_attribute("index_l")):
                index_hkl = numpy.array([refln_phase.index_h, refln_phase.index_k, refln_phase.index_l], dtype=int)
                ddict[f"index_hkl_{phase_name:}"] = index_hkl
            if refln_phase.is_attribute("f_calc"):
                f_calc = numpy.array(refln_phase.f_calc, dtype=complex)
                ddict[f"f_nucl_{phase_name:}"] = f_calc
            if (refln_phase.is_attribute("a_calc") and refln_phase.is_attribute("b_calc")):
                a_calc = numpy.array(refln_phase.a_calc, dtype=complex)
                b_calc = numpy.array(refln_phase.b_calc, dtype=complex)
                ddict[f"f_nucl_{phase_name:}"] = a_calc + 1j*b_calc

    if refln_susceptibility is not None:
        for refln_phase in refln_susceptibility:
            phase_name = refln_phase.loop_name
            if (refln_phase.is_attribute("index_h") and refln_phase.is_attribute("index_k") and refln_phase.is_attribute("index_l")):
                index_hkl = numpy.array([refln_phase.index_h, refln_phase.index_k, refln_phase.index_l], dtype=int)
                ddict[f"index_hkl_{phase_name:}"] = index_hkl
            if refln_phase.is_attribute("chi_11_calc"):
                chi_11 = numpy.array(refln_phase.chi_11_calc, dtype=complex)
                chi_12 = numpy.array(refln_phase.chi_12_calc, dtype=complex)
                chi_13 = numpy.array(refln_phase.chi_13_calc, dtype=complex)
                chi_21 = numpy.array(refln_phase.chi_21_calc, dtype=complex)
                chi_22 = numpy.array(refln_phase.chi_22_calc, dtype=complex)
                chi_23 = numpy.array(refln_phase.chi_23_calc, dtype=complex)
                chi_31 = numpy.array(refln_phase.chi_31_calc, dtype=complex)
                chi_32 = numpy.array(refln_phase.chi_32_calc, dtype=complex)
                chi_33 = numpy.array(refln_phase.chi_33_calc, dtype=complex)

                ddict[f"sft_ccs_{phase_name:}"] = numpy.stack([
                    chi_11, chi_12, chi_13, chi_21, chi_22, chi_23, chi_31, chi_32, chi_33], axis=0)

    return ddict




import numpy
from typing import NoReturn

from cryspy.A_functions_base.symmetry_constraints import calc_sc_chi
from cryspy.A_functions_base.matrix_operations import calc_m1_m2_inv_m1
from cryspy.A_functions_base.unit_cell import calc_m_m_by_unit_cell_parameters

from cryspy.A_functions_base.function_1_scat_length_neutron import \
    get_scat_length_neutron
from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

class AtomSite(ItemN):
    """Describe the atom site.

    Attributes
    ----------
        - label, type_symbol, fract_x, fract_y, fract_z (mandatory)
        - occupancy, adp_type, u_iso_or_equiv, u_equiv_geom_mean,
          b_iso_or_equiv, multiplicity, wyckoff_symbol, cartn_x, cartn_y,
          cartn_z (optional)
        - scat_length_neutron (internal)
        - space_group_wyckoff, constr_number (internal protected)
    """
    ATTR_MANDATORY_NAMES = ("label", "type_symbol", "fract_x", "fract_y",
                            "fract_z")
    ATTR_MANDATORY_TYPES = (str, str, float, float, float)
    ATTR_MANDATORY_CIF = ("label", "type_symbol", "fract_x", "fract_y",
                          "fract_z")

    ATTR_OPTIONAL_NAMES = (
        "occupancy", "adp_type",  "u_iso_or_equiv", "u_equiv_geom_mean",
        "b_iso_or_equiv", "multiplicity", "wyckoff_symbol", "cartn_x",
        "cartn_y", "cartn_z")
    ATTR_OPTIONAL_TYPES = (float, str, float, float, float, int, str, float,
                           float, float)
    ATTR_OPTIONAL_CIF = (
        "occupancy", "adp_type",  "U_iso_or_equiv", "U_equiv_geom_mean",
        "B_iso_or_equiv", "multiplicity", "Wyckoff_symbol", "Cartn_x",
        "Cartn_y", "Cartn_z")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("scat_length_neutron", )
    ATTR_INT_PROTECTED_NAMES = ("space_group_wyckoff", "constr_number", "sc_chi")

    # parameters considered are refined parameters
    ATTR_REF = ("fract_x", "fract_y", "fract_z", "occupancy", "u_iso_or_equiv",
                "u_equiv_geom_mean", "b_iso_or_equiv")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {'fract_x': "{:.6f}", 'fract_y': "{:.6f}",
                 'fract_z': "{:.6f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {"adp_type": ["Uani", "Uiso", "Uovl", "Umpe", "Bani",
                                  "Biso", "Bovl"]}

    # default values for the parameters
    D_DEFAULT = {"occupancy": 1.}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_site"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomSite, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self) -> bool:
        if self.is_attribute("space_group_wyckoff"):
            space_group_wyckoff = self.space_group_wyckoff
            self.__dict__["multiplicity"] = space_group_wyckoff.multiplicity
            self.__dict__["wyckoff_symbol"] = space_group_wyckoff.letter

            self.__dict__["fract_x_constraint"] = False
            self.__dict__["fract_y_constraint"] = False
            self.__dict__["fract_z_constraint"] = False
            xyz = numpy.array([self.fract_x, self.fract_y, self.fract_z],
                              dtype=float)

            r = space_group_wyckoff.r
            if r[0, 0] == 0:
                self.__dict__["fract_x_constraint"] = True
                self.__dict__["fract_x_refinement"] = False
            if r[1, 1] == 0:
                self.__dict__["fract_y_constraint"] = True
                self.__dict__["fract_y_refinement"] = False
            if r[2, 2] == 0:
                self.__dict__["fract_z_constraint"] = True
                self.__dict__["fract_z_refinement"] = False

            xyz_new = space_group_wyckoff.give_default_xyz(xyz)
            self.__dict__["fract_x"] = float(xyz_new[0])
            self.__dict__["fract_y"] = float(xyz_new[1])
            self.__dict__["fract_z"] = float(xyz_new[2])

        type_n = self.type_symbol
        self.__dict__["scat_length_neutron"] = get_scat_length_neutron(type_n)

    def define_space_group_wyckoff(self, space_group_wyckoff_l) -> bool:
        """Define space group by Wyckoff."""
        fract_x, fract_y, fract_z = self.fract_x, self.fract_y, self.fract_z
        x, y, z = float(fract_x), float(fract_y), float(fract_z)
        _id = space_group_wyckoff_l.get_id_for_fract(x, y, z)
        self.__dict__["space_group_wyckoff"] = space_group_wyckoff_l[_id]

    def apply_constraints(self, space_group_wyckoff_l) -> bool:
        """Apply constraints."""
        flag = not(self.is_attribute("space_group_wyckoff"))
        if flag:
            self.define_space_group_wyckoff(space_group_wyckoff_l)
        if self.is_defined():
            self.form_object()

    def calc_sc_chi(self, space_group, cell):
        """Calc sc_chi
        """
        # if self.is_attribute("sc_chi"):
        #     return self.sc_chi
        unit_cell_parameters = cell.get_unit_cell_parameters()
        x, y, z = self.fract_x, self.fract_y, self.fract_z
        o_11, o_12, o_13, o_21, o_22, o_23, o_31, o_32, o_33, o_3, o_2, o_3 =\
            space_group.calc_el_symm_for_xyz(x,y,z)
        zero_val = numpy.zeros(o_11.shape, dtype=int)
        symm_elems = numpy.array([zero_val, zero_val, zero_val, zero_val,
            o_11, o_12, o_13, o_21, o_22, o_23, o_31, o_32, o_33], dtype=int)

        sc_chi = calc_sc_chi(symm_elems, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
        self.__dict__["sc_chi"] = sc_chi
        return sc_chi

    def calc_constr_number(self, space_group):
        """
        According to table 1 in Peterse, Palm, Acta Cryst.(1966), 20, 147.
        """
        if self.is_attribute("constr_number"):
            return self.constr_number
        x, y, z = self.fract_x, self.fract_y, self.fract_z
        o_11, o_12, o_13, o_21, o_22, o_23, o_31, o_32, o_33, o_3, o_2, o_3 =\
            space_group.calc_el_symm_for_xyz(x,y,z)
        b_11, b_22, b_33, b_12, b_13, b_23 = 107, 181, 41, 7, 19, 1
        i_11 = (o_11*o_11*b_11 + o_12*o_12*b_22 + o_13*o_13*b_33 +
                o_11*o_12*b_12 + o_11*o_13*b_13 + o_12*o_13*b_23 +
                o_12*o_11*b_12 + o_13*o_11*b_13 + o_13*o_12*b_23)
        i_22 = (o_21*o_21*b_11 + o_22*o_22*b_22 + o_23*o_23*b_33 +
                o_21*o_22*b_12 + o_21*o_23*b_13 + o_22*o_23*b_23 +
                o_22*o_21*b_12 + o_23*o_21*b_13 + o_23*o_22*b_23)
        i_33 = (o_31*o_31*b_11 + o_32*o_32*b_22 + o_33*o_33*b_33 +
                o_31*o_32*b_12 + o_31*o_33*b_13 + o_32*o_33*b_23 +
                o_32*o_31*b_12 + o_33*o_31*b_13 + o_33*o_32*b_23)
        i_12 = (o_11*o_21*b_11 + o_12*o_22*b_22 + o_13*o_23*b_33 +
                o_11*o_22*b_12 + o_11*o_23*b_13 + o_12*o_23*b_23 +
                o_12*o_21*b_12 + o_13*o_21*b_13 + o_13*o_22*b_23)
        i_13 = (o_11*o_31*b_11 + o_12*o_32*b_22 + o_13*o_33*b_33 +
                o_11*o_32*b_12 + o_11*o_33*b_13 + o_12*o_33*b_23 +
                o_12*o_31*b_12 + o_13*o_31*b_13 + o_13*o_32*b_23)
        i_23 = (o_21*o_31*b_11 + o_22*o_32*b_22 + o_23*o_33*b_33 +
                o_21*o_32*b_12 + o_21*o_33*b_13 + o_22*o_33*b_23 +
                o_22*o_31*b_12 + o_23*o_31*b_13 + o_23*o_32*b_23)
        r_11, r_22, r_33, r_12, r_13, r_23 = i_11.sum(), i_22.sum(),\
            i_33.sum(), i_12.sum(), i_13.sum(), i_23.sum()
        if r_13 == 0:
            if r_23 == 0:
                if r_12 == 0:
                    if r_11 == r_22:
                        if r_22 == r_33:
                            numb = 17
                        else:
                            numb = 8
                    elif r_22 == r_33:
                        numb = 12
                    else:
                        numb = 4
                elif r_11 == r_22:
                    if r_22 == 2*r_12:
                        numb = 16
                    else:
                        numb = 5
                elif r_22 == 2*r_12:
                    numb = 14
                else:
                    numb = 2
            elif r_22 == r_33:
                numb = 9
            else:
                numb = 3
        elif r_23 == 0:
            if r_22 == 2*r_12:
                numb = 13
            else:
                numb = 1
        elif r_23 == r_13:
            if r_22 == r_33:
                numb = 18
            else:
                numb = 6
        elif r_12 == r_13:
            numb = 10
        elif r_11 == 22:
            numb = 7
        elif r_22 == r_33:
            numb = 11
        elif r_22 == 2*r_12:
            numb = 15
        else:
            numb = 0  # no constraint
        self.__dict__["constr_number"] = numb
        return numb

    def report(self) -> str:
        """Report."""
        s_out = ""
        if ((self.is_attribute("scat_length_neutron")) &
                (self.is_attribute("label"))):
            s_out = \
                f'|{self.label.rjust(10):} | {self.scat_length_neutron: .3f}|'
        return s_out

    def get_flags_atom_fract_xyz(self):
        res = numpy.array([self.fract_x_refinement, self.fract_y_refinement, self.fract_z_refinement], dtype=bool)
        return res

    def get_flags_atom_occupancy(self):
        res = self.occupancy_refinement
        return res

    def get_flags_atom_b_iso(self):
        res = False
        if not(self.is_attribute("adp_type")):
            res = False
        elif self.adp_type in ["Uiso", "Uovl", "Umpe", ]:
            res = self.u_iso_or_equiv_refinement
        elif self.adp_type in ["Biso", "Bovl", ]:
            res = self.b_iso_or_equiv_refinement
        return res


class AtomSiteL(LoopN):
    """Describe the atom sites in crystal.

    Attributes
    ----------
        - label, type_symbol, fract_x, fract_y, fract_z (mandatory)
        - occupancy, adp_type, u_iso_or_equiv, u_equiv_geom_mean,
          b_iso_or_equiv, multiplicity, wyckoff_symbol, cartn_x, cartn_y,
          cartn_z (optional)
        - scat_length_neutron (internal)
        - space_group_wyckoff, constr_number (internal protected)
    """
    ITEM_CLASS = AtomSite
    ATTR_INDEX = "label"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomSiteL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def apply_constraints(self, space_group_wyckoff_l) -> bool:
        """Apply constraints."""
        for item in self.items:
            item.apply_constraints(space_group_wyckoff_l)

    def calc_constr_number(self, space_group):
        """Calc constr number."""
        l_numb = [item.calc_constr_number(space_group) for item in self.items]
        return l_numb

    def report(self) -> str:
        """Report."""
        ls_out = ["# Scattering amplitude"]
        ls_out.append("| type| b_scat|")
        ls_out.append("|-----|-------|")
        ls_out.extend([item.report() for item in self.items])
        return "\n".join(ls_out)

    def get_flags_atom_fract_xyz(self):
        flags_atom_fract_xyz = numpy.stack([item.get_flags_atom_fract_xyz() for item in self.items], axis=0).transpose()
        return flags_atom_fract_xyz

    def get_flags_atom_occupancy(self):
        flags_atom_occupancy = numpy.stack([item.get_flags_atom_occupancy() for item in self.items], axis=0)
        return flags_atom_occupancy

    def get_flags_atom_b_iso(self):
        flags_atom_b_iso = numpy.stack([item.get_flags_atom_b_iso() for item in self.items], axis=0)
        return flags_atom_b_iso


# s_cont = """
#  loop_
#  _atom_site_label
#  _atom_site_type_symbol
#  _atom_site_fract_x
#  _atom_site_fract_y
#  _atom_site_fract_z
#  _atom_site_adp_type
#  _atom_site_B_iso_or_equiv
#  _atom_site_occupancy
#   Fe3A   Fe  0.12500 0.12500 0.12500  Uani   0.0   1.0
#   Fe3B   Fe  0.50000 0.50000 0.50000  Uani   0.0   1.0
#   O1     O   0.25521 0.25521 0.25521  Uiso   0.0   1.0
#   """

# obj = AtomSiteL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["O1"].scat_length_neutron, end="\n\n")



"""Describe AtomSiteSusceptibility, AtomSiteSusceptibilityL."""
from typing import NoReturn
import math
import numpy

from cryspy.A_functions_base.function_1_algebra import calc_m_sigma
from cryspy.A_functions_base.function_1_atomic_vibrations import \
    vibration_constraints
from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

import cryspy.A_functions_base.local_susceptibility as local_susceptibility

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_atom_site_susceptibility import calc_constr_matrix, calc_independent_lines

na = numpy.newaxis


class AtomSiteExchange(ItemN):
    """Exchange interaction (only in case of G-tensor)
    """

    ATTR_MANDATORY_NAMES = ("label", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("label", )

    ATTR_OPTIONAL_NAMES = (
        "j_type", "j_11", "j_22", "j_33", "j_12",
        "j_13", "j_23", )
    ATTR_OPTIONAL_TYPES = (str, float, float, float, float, float, float, )
    ATTR_OPTIONAL_CIF = (
        "J_type", "J_11", "J_22", "J_33", "J_12",
        "J_13", "J_23", )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("j_11", "j_22", "j_33", "j_12", "j_13", "j_23",)
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"j_11": "{:.5f}", "j_22": "{:.5f}", "j_33": "{:.5f}",
                 "j_12": "{:.5f}", "j_13": "{:.5f}", "j_23": "{:.5f}",}

    # constraints on the parameters
    D_CONSTRAINTS = {"j_type": ["Jiso", "Jani"],
                     }

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_site_exchange"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomSiteExchange, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


    def apply_space_group_constraint(self, atom_site, space_group, cell):
        """
        Space group constraints.

        According to table 1 in Peterse, Palm, Acta Cryst.(1966), 20, 147
        """
        label_aniso = self.label
        label = atom_site.label
        index = label.index(label_aniso)

        item_as = atom_site.items[index] 
        sc_exchange = item_as.calc_sc_chi(space_group, cell)
        # it should be checked
        sc_exchange = calc_constr_matrix(sc_exchange)
        

        # l_numb = atom_site.calc_constr_number(space_group)

        flag_j = self.is_attribute("j_type")
        if flag_j:
            flag_j = self.j_type.lower().startswith("jani")

        if flag_j:
            self.__dict__["j_11_constraint"] = False
            self.__dict__["j_22_constraint"] = False
            self.__dict__["j_33_constraint"] = False
            self.__dict__["j_12_constraint"] = False
            self.__dict__["j_13_constraint"] = False
            self.__dict__["j_23_constraint"] = False

        if flag_j:
            j_i = numpy.array([
                self.j_11, self.j_22, self.j_33,
                self.j_12, self.j_13, self.j_23], dtype=float)

            j_sigma_i = numpy.array([
                self.j_11_sigma, self.j_22_sigma, self.j_33_sigma,
                self.j_12_sigma, self.j_13_sigma, self.j_23_sigma], dtype=float)

            j_ref_i = numpy.array([
                self.j_11_refinement, self.j_22_refinement, self.j_33_refinement,
                self.j_12_refinement, self.j_13_refinement, self.j_23_refinement], dtype=bool)

            j_i_c = sc_exchange.dot(j_i)
            j_sigma_i_c = numpy.sqrt(numpy.square(sc_exchange).dot(numpy.square(j_sigma_i)))
            l_ind_independent = calc_independent_lines(sc_exchange)
            
            j_con_i_c = numpy.ones(j_i_c.shape, dtype=bool)
            j_con_i_c[l_ind_independent] = False
            sc_exchange_bool = numpy.logical_not(numpy.isclose(numpy.round(sc_exchange, decimals=5), 0.))
            
            j_ref_i_c = sc_exchange_bool.dot(j_ref_i) * numpy.logical_not(j_con_i_c)

            self.__dict__["j_11"], self.__dict__["j_22"], \
                self.__dict__["j_33"], self.__dict__["j_12"], \
                self.__dict__["j_13"], self.__dict__["j_23"] = \
                j_i_c[0], j_i_c[1], j_i_c[2], j_i_c[3], j_i_c[4], j_i_c[5]

            self.__dict__["j_11_sigma"], self.__dict__["j_22_sigma"], \
                self.__dict__["j_33_sigma"], self.__dict__["j_12_sigma"], \
                self.__dict__["j_13_sigma"], self.__dict__["j_23_sigma"] =\
                j_sigma_i_c[0], j_sigma_i_c[1], j_sigma_i_c[2], j_sigma_i_c[3], \
                j_sigma_i_c[4], j_sigma_i_c[5]

            self.__dict__["j_11_refinement"], \
                self.__dict__["j_22_refinement"], \
                self.__dict__["j_33_refinement"], \
                self.__dict__["j_12_refinement"], \
                self.__dict__["j_13_refinement"], \
                self.__dict__["j_23_refinement"] = j_ref_i_c[0], j_ref_i_c[1], j_ref_i_c[2],\
                    j_ref_i_c[3], j_ref_i_c[4], j_ref_i_c[5]

            self.__dict__["j_11_constraint"], \
                self.__dict__["j_22_constraint"], \
                self.__dict__["j_33_constraint"], \
                self.__dict__["j_12_constraint"], \
                self.__dict__["j_13_constraint"], \
                self.__dict__["j_23_constraint"] = j_con_i_c[0], j_con_i_c[1], j_con_i_c[2],\
                    j_con_i_c[3], j_con_i_c[4], j_con_i_c[5]
   

    def apply_j_iso_constraint(self):
        """Isotropic constraint on exchange."""
        if not(self.is_attribute("j_type")):
            return
        j_type = self.j_type
        if j_type.lower().startswith("jiso"):
            self.__dict__["j_22"] = self.j_11
            self.__dict__["j_33"] = self.j_11
            self.__dict__["j_12"] = 0. 
            self.__dict__["j_13"] = 0. 
            self.__dict__["j_23"] = 0. 
            self.__dict__["j_22_sigma"] = self.j_11_sigma
            self.__dict__["j_33_sigma"] = self.j_11_sigma
            self.__dict__["j_12_sigma"] = 0. 
            self.__dict__["j_13_sigma"] = 0. 
            self.__dict__["j_23_sigma"] = 0. 
            self.__dict__["j_22_refinement"] = False
            self.__dict__["j_33_refinement"] = False
            self.__dict__["j_12_refinement"] = False
            self.__dict__["j_13_refinement"] = False
            self.__dict__["j_23_refinement"] = False
            self.__dict__["j_22_constraint"] = True
            self.__dict__["j_33_constraint"] = True
            self.__dict__["j_12_constraint"] = True
            self.__dict__["j_13_constraint"] = True
            self.__dict__["j_23_constraint"] = True

    def get_flags_exchange(self):
        res = numpy.array([
            self.j_11_refinement, self.j_22_refinement, self.j_33_refinement,
            self.j_12_refinement, self.j_13_refinement, self.j_23_refinement], dtype=bool)
        return res



class AtomSiteExchangeL(LoopN):
    """Exchange parameters for magnetic atoms.

    Methods
    -------
        - apply_space_group_constraint
        - apply_chi_iso_constraint
    """

    ITEM_CLASS = AtomSiteExchange
    ATTR_INDEX = "label"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomSiteExchangeL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def apply_space_group_constraint(self, atom_site, space_group, cell):
        """Apply space group constraint."""
        for item in self.items:
            item.apply_space_group_constraint(atom_site, space_group, cell)

    def apply_j_iso_constraint(self):
        """Apply isotropic constraint on susceptibility."""
        for item in self.items:
            item.apply_j_iso_constraint()

    def get_flags_exchange(self):
        flags_exchange = numpy.stack([item.get_flags_exchange() for item in self.items], axis=0).transpose()
        return flags_exchange



"""DiffrnOrientMatrix and DiffrnOrientMatrixL classes."""
from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_1_matrices import\
    calc_product_matrices, calc_product_matrix_vector
from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_cell import Cell


class DiffrnOrientMatrix(ItemN):
    """UB-matrix desription.

    Data items in the DIFFRN_ORIENT_MATRIX category record details
    about the orientation matrix used in the measurement of the
    diffraction intensities.

    By default the UB matrix are given in ccsl notation

    Attributes
    ----------
        - ub_11, ub_12, ub_13, ub_21, ub_22, ub_23, ub_31, ub_32, ub_33
          (mandatory)
        - "id", "type" (optional)
        - u_11, u_12, u_13, u_21, u_22, u_23, u_31, u_32, u_33 (internal)
        - cell (protected)

    Constraints
    -----------
        - type: "CCSL", "6T2@LLB", "5C1@LLB", "BusingLevy", "-YXZ", "X-YZ",
                "XYZ"
    """

    ATTR_MANDATORY_NAMES = ("ub_11", "ub_12", "ub_13", "ub_21", "ub_22",
                            "ub_23", "ub_31", "ub_32", "ub_33")
    ATTR_MANDATORY_TYPES = (float, float, float, float, float, float,
                            float, float, float)
    ATTR_MANDATORY_CIF = ("ub_11", "ub_12", "ub_13", "ub_21", "ub_22", "ub_23",
                          "ub_31", "ub_32", "ub_33")

    ATTR_OPTIONAL_NAMES = ("id", "type")
    ATTR_OPTIONAL_TYPES = (str, str)
    ATTR_OPTIONAL_CIF = ("id", "type")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("u_11", "u_12", "u_13", "u_21", "u_22", "u_23",
                      "u_31", "u_32", "u_33", "ub", "u")
    ATTR_INT_PROTECTED_NAMES = ("cell", )

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {"type": ["CCSL", "6T2@LLB", "5C1@LLB", "BusingLevy",
                              "-YXZ", "X-YZ", "XYZ"]}

    # default values for the parameters
    D_DEFAULT = {"type": "CCSL"}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "diffrn_orient_matrix"

    def __init__(self, **kwargs) -> NoReturn:
        super(DiffrnOrientMatrix, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def __recalc_u_cell(self):

        ub_ccsl = self.calc_ub_ccsl()

        v_b_1 = ub_ccsl[:, 0]
        v_b_2 = ub_ccsl[:, 1]
        v_b_3 = ub_ccsl[:, 2]

        b_1 = float(numpy.sqrt(((v_b_1*v_b_1).sum())))
        b_2 = float(numpy.sqrt(((v_b_2*v_b_2).sum())))
        b_3 = float(numpy.sqrt(((v_b_3*v_b_3).sum())))

        c_i_g = (v_b_1*v_b_2).sum()/(b_1*b_2)
        c_i_b = (v_b_1*v_b_3).sum()/(b_1*b_3)
        c_i_a = (v_b_2*v_b_3).sum()/(b_2*b_3)

        i_a, i_b = numpy.arccos(c_i_a), numpy.arccos(c_i_b)
        i_g = numpy.arccos(c_i_g)

        s_i_a, s_i_b, s_i_g = numpy.sin(i_a), numpy.sin(i_b), numpy.sin(i_g)

        i_vol = b_1*b_2*b_3*(1-c_i_a**2-c_i_b**2-c_i_g**2+2.*c_i_a*c_i_b*c_i_g
                             )**0.5

        c_a = (c_i_b*c_i_g-c_i_a)/(s_i_b*s_i_g)
        c_b = (c_i_a*c_i_g-c_i_b)/(s_i_a*s_i_g)
        c_g = (c_i_a*c_i_b-c_i_g)/(s_i_a*s_i_b)

        a_1 = b_2*b_3*s_i_a/i_vol
        a_2 = b_1*b_3*s_i_b/i_vol
        a_3 = b_1*b_2*s_i_g/i_vol

        alpha, beta = numpy.arccos(c_a), numpy.arccos(c_b)
        gamma = numpy.arccos(c_g)

        if not(self.is_attribute("cell")):
            cell = Cell(length_a=a_1, length_b=a_2, length_c=a_3,
                        angle_alpha=float(alpha*180./numpy.pi),
                        angle_beta=float(beta*180./numpy.pi),
                        angle_gamma=float(gamma*180./numpy.pi))
            self.__dict__["cell"] = cell
        else:
            cell = self.cell
            cell.length_a, cell.length_b, cell.length_c = a_1, a_2, a_3
            cell.angle_alpha = float(alpha*180./numpy.pi)
            cell.angle_beta = float(beta*180./numpy.pi)
            cell.angle_gamma = float(gamma*180./numpy.pi)
        cell.form_object
        m_b = cell.m_b
        m_ib = numpy.linalg.inv(m_b)
        u_ccsl = numpy.dot(ub_ccsl,  m_ib)

        self.__dict__["u_11"] = float(u_ccsl[0, 0])
        self.__dict__["u_12"] = float(u_ccsl[0, 1])
        self.__dict__["u_13"] = float(u_ccsl[0, 2])
        self.__dict__["u_21"] = float(u_ccsl[1, 0])
        self.__dict__["u_22"] = float(u_ccsl[1, 1])
        self.__dict__["u_23"] = float(u_ccsl[1, 2])
        self.__dict__["u_31"] = float(u_ccsl[2, 0])
        self.__dict__["u_32"] = float(u_ccsl[2, 1])
        self.__dict__["u_33"] = float(u_ccsl[2, 2])

    def calc_ub_ccsl(self):
        """Calculate UB Matrix in CCSL notation.

        Based on the given UB matrix with defined type variable.
        """
        s_notation = self.type
        a1, a2, a3 = self.ub_11, self.ub_12, self.ub_13
        b1, b2, b3 = self.ub_21, self.ub_22, self.ub_23
        c1, c2, c3 = self.ub_31, self.ub_32, self.ub_33

        ub_ccsl = numpy.array([[a1, a2, a3],
                               [b1, b2, b3],
                               [c1, c2, c3]], dtype=float)
        if s_notation in ["XYZ", "6T2@LLB"]:
            ub_ccsl = numpy.array([[a1, a2, a3],
                                   [b1, b2, b3],
                                   [c1, c2, c3]], dtype=float)
        elif s_notation in ["-YXZ", "5C1@LLB"]:
            ub_ccsl = numpy.array([[-b1, -b2, -b3],
                                   [a1, a2, a3],
                                   [c1, c2, c3]], dtype=float)
            # ub_ccsl = numpy.array([[-b1,-b2,-b3],
            #                       [ a1, a2, a3],
            #                       [ c1, c2, c3]], dtype=float)
        elif s_notation in ["Y-XZ", "BusingLevy"]:
            ub_ccsl = numpy.array([[b1, b2, b3],
                                   [-a1, -a2, -a3],
                                   [c1, c2, c3]], dtype=float)
        return ub_ccsl

    def calc_matrix_from_ccsl(self, matrix_ccsl):
        """Calculate UB Matrix (or U Matrix).

        Calculations are done in the notation defined by type variable
        based on the UB Matrix (or U Matrix) in CCSL notation.
        """
        s_notation = self.get_notation()
        a1, a2, a3 = matrix_ccsl[0, 0], matrix_ccsl[0, 1], matrix_ccsl[0, 2]
        b1, b2, b3 = matrix_ccsl[1, 0], matrix_ccsl[1, 1], matrix_ccsl[1, 2]
        c1, c2, c3 = matrix_ccsl[2, 0], matrix_ccsl[2, 1], matrix_ccsl[2, 2]

        # the matrices are the same as in self.calc_ub_ccsl()
        # because Op. x Op. = 1
        ub = numpy.array([[a1, a2, a3],
                          [b1, b2, b3],
                          [c1, c2, c3]], dtype=float)
        if s_notation == "6T2@LLB":
            ub = numpy.array([[a1, a2, a3],
                              [b1, b2, b3],
                              [c1, c2, c3]], dtype=float)
        elif s_notation == "5C1@LLB":
            ub = numpy.array([[-b1, -b2, -b3],
                              [a1, a2, a3],
                              [c1, c2, c3]], dtype=float)
            # ub = numpy.array([[-b1, -b2, -b3],
            #                  [ a1, a2, a3],
            #                  [ c1, c2, c3]], dtype=float)
        elif s_notation == "BusingLevy":
            ub = numpy.array([[b1, b2, b3],
                              [-a1, -a2, -a3],
                              [c1, c2, c3]], dtype=float)
        return ub

    @classmethod
    def calc_ub(cls, u, type="CCSL",
                abc_angles=(6.8, 6.8, 6.8, 90., 90., 90.)):
        """Calculate UB matrix by U matrix.

        U matrix is given in the notation defined by type
        variable and cell parameters (a, b, c, alpha, beta, gamma)
        """
        (a, b, c, al, be, ga) = abc_angles[:6]
        cell = Cell(length_a=a, length_b=b, length_c=c, angle_alpha=al,
                    angle_beta=be, angle_gamma=ga)
        m_b = cell.m_b
        obj = cls(type=type)

        # in general case it should not be correct as Op. x Op.^-1 != 1
        u_ccsl = obj.calc_matrix_from_ccsl(u)
        ub_ccsl = numpy.dot(u_ccsl, m_b)
        ub = obj.calc_matrix_from_ccsl(ub_ccsl)

        obj.ub_11 = ub[0, 0]
        obj.ub_12 = ub[0, 1]
        obj.ub_13 = ub[0, 2]
        obj.ub_21 = ub[1, 0]
        obj.ub_22 = ub[1, 1]
        obj.ub_23 = ub[1, 2]
        obj.ub_31 = ub[2, 0]
        obj.ub_32 = ub[2, 1]
        obj.ub_33 = ub[2, 2]
        obj.form_object()
        return obj

    def form_object(self) -> bool:
        """Redefine parent form object."""
        flag = False
        if self.is_defined():
            ub = numpy.array([[self.ub_11, self.ub_12, self.ub_13],
                              [self.ub_21, self.ub_22, self.ub_23],
                              [self.ub_31, self.ub_32, self.ub_33]],
                             dtype=float)
            self.__dict__["ub"] = ub
            self.__recalc_u_cell()

            u = numpy.array([[self.u_11, self.u_12, self.u_13],
                             [self.u_21, self.u_22, self.u_23],
                             [self.u_31, self.u_32, self.u_33]],
                            dtype=float)
            self.__dict__["u"] = u
            flag = True
        return flag

    def calc_e_up(self, phi=0., omega=0., chi=0.):
        """
        phi, omega, chi = angles of detector in radians.

        They are zeros by defaults.
        """
        orientation_ij = (self.u_11, self.u_12, self.u_13,
                          self.u_21, self.u_22, self.u_23,
                          self.u_31, self.u_32, self.u_33)

        phi_ij = (numpy.cos(phi), numpy.sin(phi), 0.*phi,
                  -numpy.sin(phi), numpy.cos(phi), 0.*phi,
                  0.*phi, 0.*phi, 0.*phi+1.)

        omega_ij = (numpy.cos(omega), numpy.sin(omega), 0.*omega,
                    -numpy.sin(omega), numpy.cos(omega), 0.*omega,
                    0.*omega, 0.*omega, 0.*omega+1.)

        chi_ij = (numpy.cos(chi), 0.*chi, numpy.sin(chi),
                  0.*chi, 0.*chi+1., 0.*chi,
                  -numpy.sin(chi), 0.*chi, numpy.cos(chi))

        u_11, u_12, u_13, u_21, u_22, u_23, u_31, u_32, u_33 = \
            calc_product_matrices(omega_ij, chi_ij, phi_ij, orientation_ij)
        ut_ij = (u_11, u_21, u_31, u_12, u_22, u_32, u_13, u_23, u_33)
        e_up_1, e_up_2, e_up_3 = calc_product_matrix_vector(
            ut_ij, (0., 0., 1.))
        return numpy.array([e_up_1, e_up_2, e_up_3], dtype=float)

    def calc_angle(self, index_h, index_k, index_l, wavelength: float = 1.4,
            diffracted_beam: str = "left", diffractometer_axes: str = "anticlockwise"):
        """Calculate scattering angles for given reflection hkl.

        Output
        ------
            gamma  is the azimuthal angle (degrees)
            nu is the elevation angle in the laboratory coordinate system
                (xyz), where x||k_i, z||H. (degrees)
            phi is ... (degrees)
        """
        ub = self.ub
        q_ub = numpy.dot(ub, [index_h, index_k, index_l])
        q = numpy.linalg.norm(q_ub)
        gamma_0 = 0.
        nu_0 = 0.

        q_final = [- q**2 * wavelength / 2.0, 0, q_ub[2]]
        q_y2 = q**2 - q_final[0]**2 - q_final[2]**2
        flag_hh_2 = True

        if q_y2 < 0:
            # print("Angles are not found.")
            flag_hh_2 = False
        if flag_hh_2:
            q_final[1] = numpy.sqrt(q_y2)
            k_f = [q_final[0] + 1.0/wavelength, q_final[1], q_final[2]]
            nu = numpy.arcsin(k_f[2]*wavelength)/numpy.pi*180. - nu_0

            gamma = numpy.arctan2(k_f[1], k_f[0]) / numpy.pi*180. - gamma_0
            phi = (numpy.arctan2(q_final[1], q_final[0]) -
                   numpy.arctan2(q_ub[1], q_ub[0]))/numpy.pi*180
            [phi, gamma, nu] = [phi if phi > 0. else phi + 360., gamma if
                                gamma > 0. else gamma + 360., nu]
            if ((diffracted_beam.strip().lower() == "left") and
                (diffractometer_axes.strip().lower() =="anticlockwise")): # left, anticlockwise
                pass
            elif ((diffracted_beam.strip().lower() != "left") and
                (diffractometer_axes.strip().lower() != "anticlockwise")): # right, clockwise
                pass
            elif ((diffracted_beam.strip().lower() == "left") and
                (diffractometer_axes.strip().lower() !="anticlockwise")): # right, anticlockwise
                gamma = -gamma
                phi = -phi
            elif ((diffracted_beam.strip().lower() != "left") and
                (diffractometer_axes.strip().lower() == "anticlockwise")): # left, clockwise
                gamma = -gamma
                phi = -phi

            return gamma, nu, phi

    def calc_q2(self, index_h: numpy.ndarray, index_k: numpy.ndarray,
                index_l: numpy.ndarray, cell: Cell) -> numpy.ndarray:
        r"""Calculate q2 = (sin \alpha)^2.

        \alpha is angle between magnetic field and scattering vector (?)
        """
        k_1, k_2, k_3 = cell.calc_k_loc(index_h, index_k, index_l)
        h_1_loc, h_2_loc = float(self.u_31), float(self.u_32)
        h_3_loc = float(self.u_33)
        res = 1.-(k_1*h_1_loc + k_2*h_2_loc + k_3*h_3_loc)**2
        return res

    def report(self):
        s_out = f"""Orientation matrix U:
          a* [c,  a*]        c
 X: {self.u_11: 8.5f} {self.u_12: 8.5f} {self.u_13: 8.5f}
 Y: {self.u_21: 8.5f} {self.u_22: 8.5f} {self.u_23: 8.5f}
 Z: {self.u_31: 8.5f} {self.u_32: 8.5f} {self.u_33: 8.5f}

axis 'X' is along incident beam;
axis 'Z' is vertical direction."""
        return s_out

    def report_html(self):
        s_out = f"""<b>Orientation matrix U:</b><br>
        <table>
<tr><th></th><th>a*</th><th>[c,  a*]</th><th>c</th></tr>
<tr><th>X:</th><td>{self.u_11: 8.5f}</td><td>{self.u_12: 8.5f}</td><td>{self.u_13: 8.5f}</td></tr>
<tr><th>Y:</th><td>{self.u_21: 8.5f}</td><td>{self.u_22: 8.5f}</td><td>{self.u_23: 8.5f}</td></tr>
<tr><th>Z:</th><td>{self.u_31: 8.5f}</td><td>{self.u_32: 8.5f}</td><td>{self.u_33: 8.5f}</td></tr>
        </table>
axis 'X' is along incident beam;<br>
axis 'Z' is vertical direction."""
        return s_out

    def get_dictionary(self):
        res = {}
        u_matrix = self.u
        e_up = numpy.array([u_matrix[2,0], u_matrix[2,1], u_matrix[2,2]], dtype=float)

        res["matrix_u"] = numpy.array([
            self.u_11, self.u_12, self.u_13,
            self.u_21, self.u_22, self.u_23,
            self.u_31, self.u_32, self.u_33], dtype = float)
        res["matrix_ub"] = numpy.array([
            self.ub_11, self.ub_12, self.ub_13,
            self.ub_21, self.ub_22, self.ub_23,
            self.ub_31, self.ub_32, self.ub_33], dtype = float)
        return res

class DiffrnOrientMatrixL(LoopN):
    """Description of DiffrnOrientMatrixL in loop."""

    ITEM_CLASS = DiffrnOrientMatrix
    ATTR_INDEX = "id"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(DiffrnOrientMatrixL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

# s_cont = """
#  _diffrn_orient_matrix_UB_11           -0.04170
#  _diffrn_orient_matrix_UB_12           -0.01429
#  _diffrn_orient_matrix_UB_13           -0.02226
#  _diffrn_orient_matrix_UB_21           -0.00380
#  _diffrn_orient_matrix_UB_22           -0.05578
#  _diffrn_orient_matrix_UB_23           -0.05048
#  _diffrn_orient_matrix_UB_31            0.00587
#  _diffrn_orient_matrix_UB_32           -0.13766
#  _diffrn_orient_matrix_UB_33            0.02277
#   """

# obj = DiffrnOrientMatrix.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj.cell, end="\n\n")



import numpy

from cryspy.A_functions_base.unit_cell import \
    calc_volume_uc_by_unit_cell_parameters, calc_eq_ccs_by_unit_cell_parameters, \
    calc_sthovl_by_unit_cell_parameters

from cryspy.A_functions_base.structure_factor import \
    calc_f_nucl_by_dictionary, calc_sft_ccs_by_dictionary, calc_f_m_perp_by_sft

from cryspy.A_functions_base.extinction import \
    calc_extinction_sphere

from cryspy.A_functions_base.flip_ratio import \
    calc_intensities_by_structure_factors, calc_flip_ratio_by_iint, \
    calc_asymmetry_by_iint

na = numpy.newaxis



def get_flags(obj_dict):
    obj_dict_keys = obj_dict.keys()
    dict_out_variable = {}
    dict_out_refinement = {}
    for key in obj_dict_keys:
        if key.startswith("flags_"):
            label = key[key.find("_")+1:]
            flags = obj_dict[key]
            if numpy.any(flags):
                dict_out_variable[(label, )] = obj_dict[label]
                dict_out_refinement[(label, )] = flags
        if isinstance(obj_dict[key], dict):
            data_dict = obj_dict[key]
            data_dict_keys = data_dict.keys()
            for key_2 in data_dict_keys:
                if key_2.startswith("flags_"):
                    label_2 = key_2[key_2.find("_")+1:]
                    flags_2 = data_dict[key_2]
                    if numpy.any(flags_2):
                        dict_out_variable[(key, label_2)] = data_dict[label_2]
                        dict_out_refinement[(key, label_2)] = flags_2
            
    return dict_out_refinement


def calc_chi_sq_for_diffrn_by_dictionary(
        dict_diffrn, dict_crystal, dict_in_out: dict = None,
        flag_use_precalculated_data: bool=False,
        flag_calc_analytical_derivatives: bool = False):
    """Calculate chi_sq for diffrn experiment.
    """
    if dict_in_out is None:
        flag_dict = False
        flag_use_precalculated_data = False
        dict_in_out_keys = []
    else:
        flag_dict = True
        dict_in_out_keys = dict_in_out.keys()

    dict_diffrn_keys = dict_diffrn.keys()

    unit_cell_parameters = dict_crystal["unit_cell_parameters"]
    index_hkl = dict_diffrn["index_hkl"]
    flags_unit_cell_parameters = dict_crystal["flags_unit_cell_parameters"]
    flag_unit_cell_parameters = numpy.any(flags_unit_cell_parameters)
    flag_volume_unit_cell = flag_unit_cell_parameters
    volume_unit_cell, dder_vuc = calc_volume_uc_by_unit_cell_parameters(unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)
    flag_sthovl = flag_unit_cell_parameters
    magnetic_field = dict_diffrn["magnetic_field"]
    flag_magnetic_field=False

    matrix_u = dict_diffrn["matrix_u"]
    wavelength = dict_diffrn["wavelength"]
    flags_wavelength = dict_diffrn["flags_wavelength"]
    flip_ratio_excluded = dict_diffrn["flip_ratio_excluded"] # it also work for intensity

    flag_unpolarized, flag_asymmetry = False, False
    if "flip_ratio_es" in dict_diffrn_keys:
        flip_ratio_es = dict_diffrn["flip_ratio_es"]
        beam_polarization = dict_diffrn["beam_polarization"]
        flags_beam_polarization = dict_diffrn["flags_beam_polarization"]
        flipper_efficiency = dict_diffrn["flipper_efficiency"]
        flags_flipper_efficiency = dict_diffrn["flags_flipper_efficiency"]
        if "flag_asymmetry" in dict_diffrn_keys:
            flag_asymmetry = dict_diffrn["flag_asymmetry"]
    elif "intensity_es" in dict_diffrn_keys:
        intensity_es = dict_diffrn["intensity_es"]
        phase_scale = dict_diffrn["phase_scale"]
        flags_phase_scale = dict_diffrn["flags_phase_scale"]
        beam_polarization = 1.
        flipper_efficiency = 1.
        flag_unpolarized = True
        flags_beam_polarization, flags_flipper_efficiency = False, False

    if "c_lambda2" in dict_diffrn_keys:
        c_lambda2 = dict_diffrn["c_lambda2"]
        flags_c_lambda2 = dict_diffrn["flags_c_lambda2"]
        index_2hkl = 2*index_hkl
    else:
        c_lambda2 = None
        flags_c_lambda2 = False
        index_2hkl = None

    flag_eq_ccs = flag_unit_cell_parameters
    if (flag_use_precalculated_data and not(flag_eq_ccs) and ("eq_ccs" in dict_in_out_keys)):
        eq_ccs = dict_in_out["eq_ccs"]
    else:
        eq_ccs, dder_eq_ccs = calc_eq_ccs_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)
        if flag_dict:
            dict_in_out["eq_ccs"] = eq_ccs

    flags_extinction_radius, flags_extinction_mosaicity = False, False
    if "extinction_model" in dict_diffrn_keys:
        extinction_model = dict_diffrn["extinction_model"]
        extinction_radius = dict_diffrn["extinction_radius"]
        flags_extinction_radius = dict_diffrn["flags_extinction_radius"]
        extinction_mosaicity = dict_diffrn["extinction_mosaicity"]
        flags_extinction_mosaicity = dict_diffrn["flags_extinction_mosaicity"]

        flag_sthovl = flag_unit_cell_parameters
        if (flag_use_precalculated_data and not(flag_sthovl) and ("sthovl" in dict_in_out_keys)):
            sthovl = dict_in_out["sthovl"]
        else:
            sthovl, dder_sthovl = calc_sthovl_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=flag_unit_cell_parameters)
            if flag_dict:
                dict_in_out["sthovl"] = sthovl
        
        flag_cos_2theta = flag_sthovl or flags_wavelength
        if (flag_use_precalculated_data and not(flag_cos_2theta) and ("cos_2theta" in dict_in_out_keys)):
            cos_2theta = dict_in_out["cos_2theta"]
        else:        
            cos_2theta = 1.-2*numpy.square(sthovl * wavelength)
            if flag_dict:
                dict_in_out["cos_2theta"] = cos_2theta

        flag_extincton = flags_extinction_radius or flags_extinction_mosaicity or flags_wavelength

        def func_extinction(f_sq, flag_f_sq: bool = False):
            return calc_extinction_sphere(
                f_sq, extinction_radius, extinction_mosaicity, volume_unit_cell, cos_2theta, wavelength,
                extinction_model, flag_f_sq=flag_f_sq, flag_radius=flags_extinction_radius,
                flag_mosaicity=flags_extinction_mosaicity,
                flag_volume_unit_cell=flag_volume_unit_cell,
                flag_cos_2theta=flag_cos_2theta,
                flag_wavelength=flags_wavelength)
    else:
        extinction_model = ""
        extinction_radius = None
        extinction_mosaicity = None
        func_extinction = None
        flag_extincton = False
    
    if (flag_use_precalculated_data and (f"dict_in_out_hkl_{dict_crystal['name']:}" in dict_in_out_keys)):
        dict_in_out_crystal_hkl = dict_in_out[f"dict_in_out_hkl_{dict_crystal['name']:}"]
        dict_in_out_crystal_hkl["index_hkl"] = index_hkl
    else:
        dict_in_out_crystal_hkl = {"index_hkl": index_hkl}
        dict_in_out[f"dict_in_out_hkl_{dict_crystal['name']:}"] = dict_in_out_crystal_hkl

    dict_in_out[f"dict_in_out_hkl_{dict_crystal['name']:}"] = dict_in_out_crystal_hkl
    f_nucl, dder_f_nucl = calc_f_nucl_by_dictionary(dict_crystal, dict_in_out_crystal_hkl, flag_use_precalculated_data=flag_use_precalculated_data)
    flag_f_nucl = len(dder_f_nucl.keys()) > 0

    sft_ccs, dder_sft_ccs = calc_sft_ccs_by_dictionary(dict_crystal, dict_in_out_crystal_hkl, flag_use_precalculated_data=flag_use_precalculated_data)
    flag_sft_ccs = len(dder_sft_ccs.keys()) > 0

    flag_mag_atom_susceptibility = "mag_atom_susceptibility" in dder_sft_ccs.keys()

    flag_f_m_perp = flag_sft_ccs or flag_magnetic_field or flag_eq_ccs
    
    if (flag_use_precalculated_data and not(flag_f_m_perp) and "f_m_perp" in dict_in_out_keys):
        f_m_perp = dict_in_out["f_m_perp"]
    else:
        f_m_perp, dder_f_m_perp = calc_f_m_perp_by_sft(
                sft_ccs, magnetic_field, eq_ccs, flag_sft_ccs=flag_sft_ccs, flag_magnetic_field=flag_magnetic_field, flag_eq_ccs=flag_eq_ccs)
        if flag_dict:
            dict_in_out["f_m_perp"] = f_m_perp

    if c_lambda2 is not None:
        if (flag_use_precalculated_data and (f"dict_in_out_2hkl_{dict_crystal['name']:}" in dict_in_out_keys)):
            dict_in_out_crystal_2hkl = dict_in_out[f"dict_in_out_2hkl_{dict_crystal['name']:}"]
            dict_in_out_crystal_2hkl["index_hkl"] = index_2hkl
        else:
            dict_in_out_crystal_2hkl = {"index_hkl": index_2hkl}
            dict_in_out[f"dict_in_out_2hkl_{dict_crystal['name']:}"] = dict_in_out_crystal_2hkl

        f_nucl_2hkl, dder_f_nucl_2hkl = calc_f_nucl_by_dictionary(dict_crystal, dict_in_out_crystal_2hkl, flag_use_precalculated_data=flag_use_precalculated_data)
        flag_f_nucl_2hkl = len(dder_f_nucl_2hkl.keys()) > 0
        if flag_dict:
            dict_in_out["f_nucl_2hkl"] = f_nucl_2hkl

        sft_ccs_2hkl, dder_sft_ccs_2hkl = calc_sft_ccs_by_dictionary(dict_crystal, dict_in_out_crystal_2hkl, flag_use_precalculated_data=flag_use_precalculated_data)
        flag_sft_ccs_2hkl = len(dder_sft_ccs_2hkl.keys()) > 0
        if flag_dict:
            dict_in_out["sft_ccs_2hkl"] = sft_ccs_2hkl

        flag_f_m_perp_2hkl = flag_sft_ccs_2hkl or flag_magnetic_field or flag_eq_ccs
        if (flag_use_precalculated_data and not(flag_f_m_perp) and "f_m_perp_2hkl" in dict_in_out_keys):
            f_m_perp_2hkl = dict_in_out["f_m_perp_2hkl"]
        else:
            f_m_perp_2hkl, dder_f_m_perp_2hkl = calc_f_m_perp_by_sft(
                sft_ccs_2hkl, magnetic_field, eq_ccs, flag_sft_ccs=flag_sft_ccs_2hkl, flag_magnetic_field=flag_magnetic_field, flag_eq_ccs=flag_eq_ccs)
            if flag_dict:
                dict_in_out["f_m_perp_2hkl"] = f_m_perp_2hkl

    else:
        f_nucl_2hkl = None
        flag_f_nucl_2hkl = False
        f_m_perp_2hkl = None
        flag_f_m_perp_2hkl = False


    flag_flip_ratio = flag_extincton or flags_beam_polarization or flags_flipper_efficiency or \
        flag_f_nucl or flag_f_m_perp or flags_c_lambda2 or flag_f_nucl_2hkl or flag_f_m_perp_2hkl
    if (flag_use_precalculated_data and not(flag_flip_ratio) 
            and "iint_plus" in dict_in_out_keys and "iint_minus" in dict_in_out_keys):
        iint_plus = dict_in_out["iint_plus"]
        iint_minus = dict_in_out["iint_minus"]
    else:
        axis_z = matrix_u[6:9]
        iint_plus, iint_minus, dder_plus, dder_minus = calc_intensities_by_structure_factors(
            beam_polarization, flipper_efficiency, f_nucl, f_m_perp, axis_z,
            func_extinction=func_extinction,
            c_lambda2=c_lambda2, f_nucl_2hkl=f_nucl_2hkl, f_m_perp_2hkl=f_m_perp_2hkl,
            flag_beam_polarization=flags_beam_polarization, flag_flipper_efficiency=flags_flipper_efficiency,
            flag_f_nucl=flag_f_nucl, flag_f_m_perp=flag_f_m_perp,
            flag_c_lambda2=flags_c_lambda2,
            flag_f_nucl_2hkl=flag_f_nucl_2hkl, flag_f_m_perp_2hkl=flag_f_m_perp_2hkl,
            dict_in_out=dict_in_out)
        if flag_dict:
            dict_in_out["iint_plus"] = iint_plus
            dict_in_out["iint_minus"] = iint_minus

    if flag_asymmetry:
        asymmetry_e = (flip_ratio_es[0] -1.)/(flip_ratio_es[0] + 1.)
        asymmetry_s = numpy.sqrt(2.)*flip_ratio_es[1] * numpy.sqrt(numpy.square(flip_ratio_es[0]) + 1.)/numpy.square(flip_ratio_es[0] + 1.)
        exp_value = numpy.stack([asymmetry_e, asymmetry_s], axis=0)

        model_exp, dder_model_exp = calc_asymmetry_by_iint(
            iint_plus, iint_minus, c_lambda2=None, iint_2hkl=None,
            flag_iint_plus=True, flag_iint_minus=True, 
            flag_c_lambda2=False, flag_iint_2hkl=False)
    elif flag_unpolarized:
        model_exp = (iint_plus + iint_minus)*phase_scale
        dict_in_out["intensity_calc"] = model_exp
        exp_value = intensity_es
        dder_model_exp = {
            "iint_plus": numpy.ones_like(iint_plus)*phase_scale,
            "iint_minus": numpy.ones_like(iint_minus)*phase_scale,
            }
    else:
        model_exp, dder_model_exp = calc_flip_ratio_by_iint(
            iint_plus, iint_minus, c_lambda2=None, iint_2hkl=None,
            flag_iint_plus=True, flag_iint_minus=True, 
            flag_c_lambda2=False, flag_iint_2hkl=False)
        exp_value = flip_ratio_es

    index_true = numpy.logical_not(flip_ratio_excluded)
    chi_sq = numpy.square((exp_value[0, index_true]-model_exp[index_true])/exp_value[1, index_true]).sum(axis=0)
    der_model_int_plus = dder_model_exp["iint_plus"]
    der_model_int_minus = dder_model_exp["iint_minus"]

    n_hkl = index_true.sum()
    if flag_dict:
        dict_in_out["chi_sq"] = chi_sq
        dict_in_out["n_hkl"] = n_hkl

    diff_exp_model = (exp_value[0, index_true]-model_exp[index_true])
    inv_sigma_sq_exp_value = 1./numpy.square(exp_value[1, index_true])
    
    dder_plus_crystal, dder_minus_crystal = {}, {}
    if flag_mag_atom_susceptibility:
        dder_f_m_perp_mas = (
            dder_f_m_perp['sft_ccs_real'][:,:, na, :, na]*dder_sft_ccs["mag_atom_susceptibility"].real[na, :, :, : , :]+
            dder_f_m_perp['sft_ccs_imag'][:,:, na, :, na]*dder_sft_ccs["mag_atom_susceptibility"].imag[na, :, :, : , :]).sum(axis=1)
        dder_plus_mas = (
            dder_plus["f_m_perp_real"][:, na, :, na] * dder_f_m_perp_mas.real + 
            dder_plus["f_m_perp_imag"][:, na, :, na] * dder_f_m_perp_mas.imag).sum(axis=0)
        dder_minus_mas = (
            dder_minus["f_m_perp_real"][:, na, :, na] * dder_f_m_perp_mas.real + 
            dder_minus["f_m_perp_imag"][:, na, :, na] * dder_f_m_perp_mas.imag).sum(axis=0)
        if c_lambda2 is not None:
            dder_f_m_perp_mas_2hkl = (
                dder_f_m_perp_2hkl['sft_ccs_real'][:,:, na, :, na]*dder_sft_ccs_2hkl["mag_atom_susceptibility"].real[na, :, :, : , :]+
                dder_f_m_perp_2hkl['sft_ccs_imag'][:,:, na, :, na]*dder_sft_ccs_2hkl["mag_atom_susceptibility"].imag[na, :, :, : , :]).sum(axis=1)
            dder_plus_mas += (
            dder_plus["f_m_perp_2hkl_real"][:, na, :, na] * dder_f_m_perp_mas_2hkl.real + 
            dder_plus["f_m_perp_2hkl_imag"][:, na, :, na] * dder_f_m_perp_mas_2hkl.imag).sum(axis=0)
            dder_minus_mas += (
            dder_minus["f_m_perp_2hkl_real"][:, na, :, na] * dder_f_m_perp_mas_2hkl.real + 
            dder_minus["f_m_perp_2hkl_imag"][:, na, :, na] * dder_f_m_perp_mas_2hkl.imag).sum(axis=0)
        dder_plus_crystal["mag_atom_susceptibility"] = numpy.swapaxes(dder_plus_mas, 0, 1)
        dder_minus_crystal["mag_atom_susceptibility"] = numpy.swapaxes(dder_minus_mas, 0, 1)
        
    dder_plus_diffrn, dder_minus_diffrn = {}, {}
    if flags_beam_polarization:
        dder_plus_diffrn["beam_polarization"] = dder_plus["beam_polarization"][:, na]
        dder_minus_diffrn["beam_polarization"] = dder_minus["beam_polarization"][:, na]
    if flags_flipper_efficiency:
        dder_minus_diffrn["flipper_efficiency"] = dder_minus["flipper_efficiency"][:, na]
    if flags_extinction_radius:
        dder_plus_diffrn["extinction_radius"] = dder_plus["radius"][:, na]
        dder_minus_diffrn["extinction_radius"] = dder_minus["radius"][:, na]
    if flags_extinction_mosaicity:
        dder_plus_diffrn["extinction_mosaicity"] = dder_plus["mosaicity"][:, na]
        dder_minus_diffrn["extinction_mosaicity"] = dder_minus["mosaicity"][:, na]
    if flags_c_lambda2:
        dder_plus_diffrn["c_lambda2"] = dder_plus["c_lambda2"][:, na]
        dder_minus_diffrn["c_lambda2"] = dder_minus["c_lambda2"][:, na]

    dder_plus_diffrn_keys = dder_plus_diffrn.keys()
    dder_minus_diffrn_keys = dder_minus_diffrn.keys()
    dder_plus_crystal_keys = dder_plus_crystal.keys()
    dder_minus_crystal_keys = dder_minus_crystal.keys()
    l_dder_plus_p = []
    l_dder_minus_p = []
    l_parameter_name = []
    der_chi_sq, dder_chi_sq = numpy.array([], dtype=float), numpy.array([[]], dtype=float)
    if True: # len(dder_plus_diffrn_keys) + len(dder_minus_diffrn_keys) + len(dder_plus_crystal_keys) + len(dder_minus_crystal_keys) > 0
        # Cacluation first and second derivatives over refinement parameters
        diffrn_type_name = dict_diffrn["type_name"]
        crystal_type_name = dict_crystal["type_name"]
        flags_diffrn = get_flags(dict_diffrn) 
        flags_crystal = get_flags(dict_crystal) 

        for way, flags in flags_diffrn.items():
            ind_1d = numpy.atleast_1d(numpy.argwhere(flags)) #.flatten()
            name = way[0]
            if ((name in dder_plus_diffrn_keys) and (name in dder_minus_diffrn_keys)):
                dder_plus_p = dder_plus_diffrn[name][:, flags] # numpy.take(dder_plus_diffrn[name], ind_1d, axis=1)
                dder_minus_p = dder_minus_diffrn[name][:, flags]
            elif name in dder_plus_diffrn_keys:
                dder_plus_p = dder_plus_diffrn[name][:, flags]
                dder_minus_p = numpy.zeros_like(dder_plus_p)
            elif name in dder_minus_diffrn_keys:
                dder_minus_p = dder_minus_diffrn[name][:, flags]
                dder_plus_p = numpy.zeros_like(dder_minus_p)
            elif name == "phase_scale":#FIXME
                dder_plus_p = numpy.zeros(shape=(index_true.size, 1), dtype=float)
                dder_minus_p = numpy.zeros(shape=(index_true.size, 1), dtype=float)
            else:
                raise AttributeError("It should not be like this.")
            parameter_name = [(diffrn_type_name, ) + way + (tuple(ind_1d[ind,:]), ) for ind in range(ind_1d.shape[0])]
            l_dder_plus_p.append(dder_plus_p)
            l_dder_minus_p.append(dder_minus_p)
            l_parameter_name.extend(parameter_name)

        for way, flags in flags_crystal.items():
            flag_der_hh = True
            ind_1d = numpy.atleast_1d(numpy.argwhere(flags)) #.flatten()
            name = way[0]
            if ((name in dder_plus_crystal_keys) and (name in dder_minus_crystal_keys)):
                dder_plus_p = dder_plus_crystal[name][:, flags]
                dder_minus_p = dder_minus_crystal[name][:, flags]
            elif name in dder_plus_crystal_keys:
                dder_plus_p = dder_plus_crystal[name][:, flags]
                dder_minus_p = numpy.zeros_like(dder_plus_p)
            elif name in dder_minus_crystal_keys:
                dder_minus_p = dder_minus_crystal[name][:, flags]
                dder_plus_p = numpy.zeros_like(dder_minus_p)
            else:
                flag_der_hh = False # raise AttributeError("It should not be like this.")

            parameter_name = [(crystal_type_name, ) + way + (tuple(ind_1d[ind,:]), ) for ind in range(ind_1d.shape[0])]
            l_parameter_name.extend(parameter_name)
            if flag_der_hh:
                l_dder_plus_p.append(dder_plus_p)
                l_dder_minus_p.append(dder_minus_p)

        if len(l_dder_plus_p) > 0:
            dder_plus_p = numpy.concatenate(l_dder_plus_p, axis=1)
            dder_minus_p = numpy.concatenate(l_dder_minus_p, axis=1)

            der_model_p = der_model_int_plus[:, na]*dder_plus_p + der_model_int_minus[:, na]*dder_minus_p
            der_chi_sq = -2.* (diff_exp_model[:, na] * inv_sigma_sq_exp_value[:, na] * der_model_p).sum(axis=0)
            dder_chi_sq = 2.* (inv_sigma_sq_exp_value[:, na, na]*(der_model_p[:, na, :]*der_model_p[:, :, na])).sum(axis=0) #w_diff is equal to zero

            # iint_minus_sq = numpy.square(iint_minus)
            # der_fr_p = (iint_minus[:, na] * dder_plus_p - iint_plus[:, na] * dder_minus_p) / iint_minus_sq[:, na]
            
            # dder_fr_p = (2*iint_minus[:, na, na] * dder_minus_p[:, :, na] * dder_minus_p[:, na, :] - 
            #    iint_minus_sq[:, na, na] * dder_plus_p[:, :, na] * dder_minus_p[:, na, :] - 
            #    dder_minus_p[:, :, na] * dder_plus_p[:, na, :]) / numpy.square(iint_minus_sq)[:, na, na]
    
            # der_chi_sq = -2.* (diff_exp_model[:, na] * inv_sigma_sq_exp_value[:, na] * der_fr_p).sum(axis=0)
            # dder_chi_sq = 2.* (inv_sigma_sq_exp_value[:, na, na]*(der_fr_p[:, na, :]*der_fr_p[:, :, na] - 
            #     0*diff_exp_model[:, na, na]*dder_fr_p)).sum(axis=0) #w_diff is equal to zero

    return chi_sq, n_hkl, der_chi_sq, dder_chi_sq, l_parameter_name
    


import numpy
import scipy
import scipy.special

na = numpy.newaxis



def calc_spectrum(time, spectrum_type, spectrum_parameters, flag_spectrum_parameters=False):
    exp = numpy.exp
    time_sq = numpy.square(time)
    time_4 = numpy.square(time_sq)
    a0 = spectrum_parameters[0]
    a1 = spectrum_parameters[1]
    a2 = spectrum_parameters[2]
    a3 = spectrum_parameters[3]
    a4 = spectrum_parameters[4]
    a5 = spectrum_parameters[5]
    a6 = spectrum_parameters[6]
    a7 = spectrum_parameters[7]
    a8 = spectrum_parameters[8]
    if spectrum_type == "Empirical-Exponents":
        res = a0 + a1 * exp(-a2 * time) + \
            a3 * exp(-a4 * time_sq) + \
            a5 * exp(-a6 * time*time_sq) + \
            a7 * exp(-a8 * time_4)
    else:  # spectrum_type == "Maxwell"
        res = a0 + a1 * exp(-a2 * time_sq)/(time*time_4) + \
            a3 * exp(-a4 * time_sq) + \
            a5 * exp(-a6 * time*time_sq) + \
            a7 * exp(-a8 * numpy.square(time_sq))
    dder = {}
    if flag_spectrum_parameters:
        pass
    return res, dder


def calc_time_for_epithermal_neutrons_by_d(d, zero, dtt1, zerot, dtt1t, dtt2t, width, x_cross):
    time_e = zero + dtt1 * d
    time_t = zerot + dtt1t * d - dtt2t / d
    n_cross = 0.5*scipy.special.erfc(width * (x_cross - 1./d))
    time = n_cross * time_e + (1-n_cross) * time_t
    return time


def calc_time_for_thermal_neutrons_by_d(d, zero, dtt1, dtt2):
    time = zero + dtt1 * d + dtt2 * numpy.square(d)
    return time


def calc_d_by_time_for_thermal_neutrons(time, zero, dtt1, dtt2):
    det = numpy.sqrt(numpy.square(dtt1) - 4.*(zero-time)*dtt2)
    if dtt2 < 0.:
        d = (det-dtt1)/(2.*dtt2)
    elif dtt2 > 0.:
        d = (-det-dtt1)/(2.*dtt2)
    else:
        d = (time - zero)/dtt1
    return d


def calc_d_min_max_by_time_thermal_neutrons(time, zero, dtt1, dtt2):
    time_min = numpy.min(time)
    time_max = numpy.max(time)
    det_sq_min = numpy.square(dtt1) - 4.* dtt2*(zero - time_min)
    det_sq_max = numpy.square(dtt1) - 4.* dtt2*(zero - time_max)
    d_max = (-dtt1+numpy.sqrt(det_sq_max))/(2.*dtt2)
    d_min = (-dtt1+numpy.sqrt(det_sq_min))/(2.*dtt2)
    return numpy.stack([d_min, d_max], axis=0)

def calc_d_min_max_by_time_epithermal_neutrons(time, zero, dtt1, zerot, dtt1t, dtt2t):
    d_min_max_t = calc_d_min_max_by_time_thermal_neutrons(time, zerot, dtt1t, dtt2t)
    time_min = numpy.min(time)
    time_max = numpy.max(time)
    d_min_max_e = numpy.stack([(time_min-zero)/dtt1, (time_max-zero)/dtt1], axis=0)
    d_min_max = numpy.array([
        min([d_min_max_t[0], d_min_max_e[0]]),
        max([d_min_max_t[1], d_min_max_e[1]])], dtype=float)
    return d_min_max


def calc_hpv_eta(h_g, h_l):
    """pseudo-Voight function
    
    calculate h_pV and eta based on Gauss and Lorentz Size
    """
    h_g_2, h_l_2 = h_g**2, h_l**2
    h_g_3, h_l_3 = h_g_2*h_g, h_l_2*h_l
    h_g_4, h_l_4 = h_g_3*h_g, h_l_3*h_l
    h_g_5, h_l_5 = h_g_4*h_g, h_l_4*h_l
    c_2, c_3, c_4, c_5 = 2.69269, 2.42843, 4.47163, 0.07842
    h_pv = (h_g_5 + c_2*h_g_4*h_l + c_3*h_g_3*h_l_2 + 
            c_4*h_g_2*h_l_3 + c_5*h_g*h_l_4 + h_l_5)**0.2
    hh = h_l*1./h_pv
    eta = 1.36603*hh - 0.47719*hh**2 + 0.11116*hh**3
    return h_pv, eta


def tof_Jorgensen(alpha, beta, sigma, time, time_hkl):
    norm = 0.5*alpha*beta/(alpha+beta)
    time_2d, time_hkl_2d = numpy.meshgrid(time, time_hkl, indexing="ij")
    delta_2d = time_2d-time_hkl_2d
    y, z, u, v = calc_y_z_u_v(alpha, beta, sigma, delta_2d)

    exp_u = exp(u)
    exp_v = exp(v)
    exp_u[numpy.isinf(exp_u)] = 1e200
    exp_v[numpy.isinf(exp_v)] = 1e200

    profile_g_2d = norm[:, na] * (exp_u * erfc(y) + exp_v * erfc(z))
    return profile_g_2d


def tof_Jorgensen_VonDreele(alpha, beta, sigma, gamma, time, time_hkl):
    two_over_pi = 2.*numpy.pi
    norm = 0.5*alpha*beta/(alpha+beta)
    time_2d, time_hkl_2d = numpy.meshgrid(time, time_hkl, indexing="ij")
    delta_2d = time_2d-time_hkl_2d

    # FIXME: it has to be checked
    # sigma = gamma*(inv_8ln2)**0.5
    h_pv, eta = calc_hpv_eta(sigma, gamma)

    y, z, u, v = calc_y_z_u_v(alpha, beta, sigma, delta_2d)

    exp_u = exp(u)
    exp_v = exp(v)
    exp_u[numpy.isinf(exp_u)] = 1e200
    exp_v[numpy.isinf(exp_v)] = 1e200

    profile_g_2d = norm[:, na] * (exp_u * erfc(y) + exp_v * erfc(z))

    z1_2d = alpha[:, na]*delta_2d + (1j*0.5*alpha*gamma)[:, na]
    z2_2d = -beta[:, na]*delta_2d + (1j*0.5*beta*gamma)[:, na]

    fz1_2d = exp1(z1_2d)
    fz2_2d = exp1(z2_2d)

    # FIXME: check it
    fz1_2d[numpy.isnan(fz1_2d)] = 0.
    fz1_2d[numpy.isinf(fz1_2d)] = 0.
    fz2_2d[numpy.isnan(fz2_2d)] = 0.
    fz2_2d[numpy.isinf(fz2_2d)] = 0.

    oml_a_2d = -imag(fz1_2d) * two_over_pi
    oml_b_2d = -imag(fz2_2d) * two_over_pi
    profile_l_2d = norm[:, na] * (oml_a_2d + oml_b_2d)
    one_e = 1. - eta
    tof_peak_2d = one_e[:, na] * profile_g_2d + eta[:, na] * profile_l_2d
    return tof_peak_2d



inv_8ln2 = 0.18033688011112042591999058512524

na = numpy.newaxis
square = numpy.square
exp = numpy.exp
imag = numpy.imag
erfc = scipy.special.erfc
exp1 = scipy.special.exp1  # E1 = \int_{z}^{\inf} exp(-t)/t dt



def calc_alpha(alpha0, alpha1, d):
    """alpha0+alpha1/d
    """
    return alpha0+alpha1/d


def calc_beta(beta0, beta1, d):
    """beta0+beta1/d**4
    """
    return beta0+beta1*numpy.power(d, -4)


def calc_sigma(d, sigma0, sigma1, sigma2, size_g:float=0., strain_g:float=0.):
    """
    H_G**2 = (sigma2+size_g)*d**4 + (sigma1+strain_g)*d**2 + sigma0
    """
    d_sq = numpy.square(d)
    d_sq_sq = numpy.square(d_sq)
    h_g_sq = numpy.abs((sigma2+size_g) * d_sq_sq + (sigma1+strain_g) * d_sq + sigma0)
    h_g = numpy.sqrt(h_g_sq)
    return h_g


def calc_sigma_gamma(
        d, sigma0, sigma1, sigma2, gamma0, gamma1, gamma2,
        size_g: float = 0., size_l: float = 0., strain_g: float = 0.,
        strain_l: float = 0.):
    """Calculate H_G (sigma) and H_L (gamma)

        H_G**2 = (sigma2+size_g)*d**4 + (sigma1+strain_g)*d**2 + sigma0
        H_L = (gamma2+size_l)*d**2 + (sigma1+strain_l)*d + sigma0
    """
    d_sq = numpy.square(d)
    h_g = calc_sigma(d, sigma0, sigma1, sigma2, size_g, strain_g)
    h_l = (gamma2+size_l) * d_sq + (gamma1+strain_l) * d + gamma0
    return h_g, h_l


def calc_y_z_u_v(alpha, beta, sigma, delta_2d):
    """Calculate y, z, u, v

    y = (alpha * sigma**2 + delta)/(sigma * 2**0.5)
    z = (beta * sigma**2 - delta)/(sigma * 2**0.5)
    u = 0.5 * alpha * (alpha*sigma**2 + 2 delta)
    v = 0.5 * beta * (beta*sigma**2 - 2 delta)

    """
    sigma_sq = square(sigma)

    y = (alpha * sigma/(2.**0.5))[:, na] + delta_2d/(sigma*2.**0.5)[:, na]
    z = (beta * sigma/(2.**0.5))[:, na] - delta_2d/(sigma*2.**0.5)[:, na]
    u = (0.5*square(alpha)*sigma_sq)[:, na] + delta_2d*alpha[:, na]
    v = (0.5*square(beta)*sigma_sq)[:, na] - delta_2d*beta[:, na]
    return y, z, u, v


def tof_Carpenter():
    pass



def calc_peak_shape_function(alphas, betas, sigmas,
        d, time, time_hkl, gammas = None, size_g: float = 0., strain_g: float = 0.,
        size_l: float = 0., strain_l: float = 0., peak_shape: str = "pseudo-Voigt"):
    """Calculate peak-shape function F(DELTA)
    For Gauss peak-shape:
        F(DELTA) = alpha * beta / (alpha+beta) * [exp(u) erfc(y) + exp(v) erfc(z)]
    For pseudo-Voigt peak-shape:
        F(DELTA) = ...
    alpha = alpha0 + alpha1 / d
    beta = beta0 + beta1 / d**4
    sigma = sigma0 + sigma1 * d
    u = alpha/2 * (alpha * sigma**2 + 2 DELTA)
    v = beta/2 * (beta * sigma**2 - 2 DELTA)
    y = (alpha * sigma**2 + DELTA)/(sigma * 2**0.5)
    z = (beta * sigma**2 - DELTA)/(sigma * 2**0.5)
    DELTA = time - time_hkl
    """
    alpha0, alpha1 = alphas[0], alphas[1]
    beta0, beta1 = betas[0], betas[1]
    alpha = alpha0 + alpha1 / d
    beta = beta0 + beta1 / d**4

    if peak_shape == "Gauss":
        sigma0, sigma1, sigma2 = sigmas[0], sigmas[1], sigmas[2]
        sigma = calc_sigma(d, sigma0, sigma1, sigma2, size_g=size_g, strain_g=strain_g)
        res_2d = tof_Jorgensen(alpha, beta, sigma, time, time_hkl)
    else:  # peak_shape == "pseudo-Voigt"
        sigma0, sigma1, sigma2 = sigmas[0], sigmas[1], sigmas[2]
        gamma0, gamma1, gamma2 = gammas[0], gammas[1], gammas[2]
        sigma, gamma = calc_sigma_gamma(
            d, sigma0, sigma1, sigma2, gamma0,
            gamma1, gamma2, size_g=size_g, size_l=size_l,
            strain_g=strain_g, strain_l=strain_l)

        res_2d = tof_Jorgensen_VonDreele(
            alpha, beta, sigma, gamma, time, time_hkl)
    return res_2d


from typing import NoReturn
import numpy
import matplotlib
import matplotlib.pyplot as plt


from cryspy.A_functions_base.powder_diffraction_const_wavelength import \
    calc_ttheta_phi_by_gamma_nu

from cryspy.A_functions_base.function_1_gamma_nu import \
    recal_int_to_gammanu_grid
    
from cryspy.B_parent_classes.cl_1_item import ItemN

na = numpy.newaxis

class Pd2dMeas(ItemN):
    """
    Measured 2D diffraction pattern (polarized neutrons).
    
    This section contains the measured diffractogram and information
    about the conditions used for the measurement of the diffraction 
    data set, prior to processing and application of correction
    terms.

    """
    ATTR_MANDATORY_NAMES = ()
    ATTR_MANDATORY_TYPES = ()
    # ("matrix", "matrix", "matrix", "matrix")
    ATTR_MANDATORY_CIF = ()

    ATTR_OPTIONAL_NAMES = (
        "ttheta_phi_intensity_plus", "ttheta_phi_intensity_plus_sigma",
        "ttheta_phi_intensity_minus", "ttheta_phi_intensity_minus_sigma",
        "gamma_nu_intensity_plus", "gamma_nu_intensity_plus_sigma",
        "gamma_nu_intensity_minus", "gamma_nu_intensity_minus_sigma",
        "gamma_nu_intensity", "gamma_nu_intensity_sigma")
    ATTR_OPTIONAL_TYPES = (
        str, str, str, str, str, str, str, str, str, str)
    ATTR_OPTIONAL_CIF = (
        "2theta_phi_intensity_plus", "2theta_phi_intensity_plus_sigma",
        "2theta_phi_intensity_minus", "2theta_phi_intensity_minus_sigma",
        "gamma_nu_intensity_plus", "gamma_nu_intensity_plus_sigma",
        "gamma_nu_intensity_minus", "gamma_nu_intensity_minus_sigma",
        "gamma_nu_intensity", "gamma_nu_intensity_sigma")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("ttheta", "phi", "intensity_plus", "intensity_plus_sigma",
                      "intensity_minus", "intensity_minus_sigma", "gamma", "nu",
                      "gn_intensity_plus", "gn_intensity_plus_sigma",
                      "gn_intensity_minus", "gn_intensity_minus_sigma",
                      "gn_intensity", "gn_intensity_sigma")
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd2d_meas"

    def __init__(self, **kwargs) -> NoReturn:
        super(Pd2dMeas, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self) -> NoReturn:
        """Redefine form_object."""
        flag_tth_phi = (self.is_attribute("ttheta_phi_intensity_plus") &
                self.is_attribute("ttheta_phi_intensity_plus_sigma") &
                self.is_attribute("ttheta_phi_intensity_minus") &
                self.is_attribute("ttheta_phi_intensity_minus_sigma"))
        flag_gamma_nu = (
                self.is_attribute("gamma_nu_intensity_plus") &
                self.is_attribute("gamma_nu_intensity_plus_sigma") &
                self.is_attribute("gamma_nu_intensity_minus") &
                self.is_attribute("gamma_nu_intensity_minus_sigma"))
        flag_gamma_nu_unpolarized = (
                self.is_attribute("gamma_nu_intensity") &
                self.is_attribute("gamma_nu_intensity_sigma"))

        if flag_tth_phi:
            l_1 = (self.ttheta_phi_intensity_plus).strip().split("\n")
            l_2 = (self.ttheta_phi_intensity_plus_sigma).strip().split("\n")
            l_3 = (self.ttheta_phi_intensity_minus).strip().split("\n")
            l_4 = (self.ttheta_phi_intensity_minus_sigma).strip().split("\n")

            l_ttheta = numpy.array([_ for _ in l_1[0].strip().split()[1:]],
                                   dtype=float)
            l_phi, ll_intensity_plus, ll_intensity_plus_sigma = [], [], []
            ll_intensity_minus, ll_intensity_minus_sigma = [], []
            for line_1, line_2, line_3, line_4 in zip(l_1[1:], l_2[1:], l_3[1:],
                                                      l_4[1:]):
                _l_1 = line_1.strip().split()
                _l_2 = line_2.strip().split()
                _l_3 = line_3.strip().split()
                _l_4 = line_4.strip().split()
                l_phi.append(float(_l_1[0]))
                ll_intensity_plus.append(_l_1[1:])
                ll_intensity_plus_sigma.append(_l_2[1:])
                ll_intensity_minus.append(_l_3[1:])
                ll_intensity_minus_sigma.append(_l_4[1:])

            ll_intensity_plus = numpy.array(ll_intensity_plus, dtype=float).transpose()
            ll_intensity_plus_sigma = numpy.array(ll_intensity_plus_sigma, dtype=float).transpose()
            ll_intensity_minus = numpy.array(ll_intensity_minus, dtype=float).transpose()
            ll_intensity_minus_sigma = numpy.array(ll_intensity_minus_sigma, dtype=float).transpose()

            self.__dict__["ttheta"] = l_ttheta
            self.__dict__["phi"] = numpy.array(l_phi, dtype=float)
            self.__dict__["intensity_plus"] = ll_intensity_plus
            self.__dict__["intensity_plus_sigma"] = ll_intensity_plus_sigma
            self.__dict__["intensity_minus"] = ll_intensity_minus
            self.__dict__["intensity_minus_sigma"] = ll_intensity_minus_sigma

        if flag_gamma_nu:
            l_1 = (self.gamma_nu_intensity_plus).strip().split("\n")
            l_2 = (self.gamma_nu_intensity_plus_sigma).strip().split("\n")
            l_3 = (self.gamma_nu_intensity_minus).strip().split("\n")
            l_4 = (self.gamma_nu_intensity_minus_sigma).strip().split("\n")

            l_gamma = numpy.array([_ for _ in l_1[0].strip().split()[1:]],
                                   dtype=float)
            l_nu, ll_intensity_plus, ll_intensity_plus_sigma = [], [], []
            ll_intensity_minus, ll_intensity_minus_sigma = [], []
            for line_1, line_2, line_3, line_4 in zip(l_1[1:], l_2[1:], l_3[1:],
                                                      l_4[1:]):
                _l_1 = line_1.strip().split()
                _l_2 = line_2.strip().split()
                _l_3 = line_3.strip().split()
                _l_4 = line_4.strip().split()
                l_nu.append(float(_l_1[0]))
                ll_intensity_plus.append(_l_1[1:])
                ll_intensity_plus_sigma.append(_l_2[1:])
                ll_intensity_minus.append(_l_3[1:])
                ll_intensity_minus_sigma.append(_l_4[1:])

            ll_intensity_plus = numpy.array(ll_intensity_plus, dtype=float).transpose()
            ll_intensity_plus_sigma = numpy.array(ll_intensity_plus_sigma, dtype=float).transpose()
            ll_intensity_minus = numpy.array(ll_intensity_minus, dtype=float).transpose()
            ll_intensity_minus_sigma = numpy.array(ll_intensity_minus_sigma, dtype=float).transpose()

            self.__dict__["gamma"] = l_gamma
            self.__dict__["nu"] = numpy.array(l_nu, dtype=float)
            self.__dict__["gn_intensity_plus"] = ll_intensity_plus
            self.__dict__["gn_intensity_plus_sigma"] = ll_intensity_plus_sigma
            self.__dict__["gn_intensity_minus"] = ll_intensity_minus
            self.__dict__["gn_intensity_minus_sigma"] = ll_intensity_minus_sigma

        if flag_gamma_nu_unpolarized:
            l_1 = (self.gamma_nu_intensity).strip().split("\n")
            l_2 = (self.gamma_nu_intensity_sigma).strip().split("\n")

            l_gamma = numpy.array([_ for _ in l_1[0].strip().split()[1:]],
                                   dtype=float)
            l_nu, ll_intensity, ll_intensity_sigma = [], [], []
            for line_1, line_2 in zip(l_1[1:], l_2[1:]):
                _l_1 = line_1.strip().split()
                _l_2 = line_2.strip().split()
                l_nu.append(float(_l_1[0]))
                ll_intensity.append(_l_1[1:])
                ll_intensity_sigma.append(_l_2[1:])

            ll_intensity = numpy.array(ll_intensity, dtype=float).transpose()
            ll_intensity_sigma = numpy.array(ll_intensity_sigma, dtype=float).transpose()

            self.__dict__["gamma"] = l_gamma
            self.__dict__["nu"] = numpy.array(l_nu, dtype=float)
            self.__dict__["gn_intensity"] = ll_intensity
            self.__dict__["gn_intensity_sigma"] = ll_intensity_sigma

    def form_ttheta_phi_intensity_plus(self) -> bool:
        if (self.is_attribute("phi") & self.is_attribute("ttheta") &
            self.is_attribute("intensity_plus")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            ll_intensity = self.intensity_plus
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for phi, l_intensity in zip(self.phi, ll_intensity):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["ttheta_phi_intensity_plus"] = "\n".join(ls_out)

    def form_ttheta_phi_intensity_plus_sigma(self) -> bool:
        if (self.is_attribute("phi") & self.is_attribute("ttheta") &
            self.is_attribute("intensity_plus_sigma")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            ll_intensity = self.intensity_plus_sigma
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for phi, l_intensity in zip(self.phi, ll_intensity):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["ttheta_phi_intensity_plus_sigma"] = "\n".join(ls_out)

    def form_ttheta_phi_intensity_minus(self) -> bool:
        if (self.is_attribute("phi") & self.is_attribute("ttheta") &
            self.is_attribute("intensity_minus")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            ll_intensity = self.intensity_minus
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for phi, l_intensity in zip(self.phi, ll_intensity):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["ttheta_phi_intensity_minus"] = "\n".join(ls_out)

    def form_ttheta_phi_intensity_minus_sigma(self) -> bool:
        if (self.is_attribute("phi") & self.is_attribute("ttheta") &
            self.is_attribute("intensity_minus_sigma")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            ll_intensity = self.intensity_minus_sigma
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for phi, l_intensity in zip(self.phi, ll_intensity):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["ttheta_phi_intensity_minus_sigma"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_plus(self) -> bool:
        if (self.is_attribute("nu") & self.is_attribute("gamma") &
            self.is_attribute("gn_intensity_plus")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.gn_intensity_plus
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_plus"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_plus_sigma(self) -> bool:
        if (self.is_attribute("nu") & self.is_attribute("gamma") &
            self.is_attribute("gn_intensity_plus_sigma")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.gn_intensity_plus_sigma
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_plus_sigma"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_minus(self) -> bool:
        if (self.is_attribute("nu") & self.is_attribute("gamma") &
            self.is_attribute("gn_intensity_minus")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.gn_intensity_minus
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_minus"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_minus_sigma(self) -> bool:
        if (self.is_attribute("nu") & self.is_attribute("gamma") &
            self.is_attribute("gn_intensity_minus_sigma")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.gn_intensity_minus_sigma
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_minus_sigma"] = "\n".join(ls_out)


    def form_gamma_nu_intensity(self) -> bool:
        if (self.is_attribute("nu") & self.is_attribute("gamma") &
            self.is_attribute("gn_intensity")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.gn_intensity
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_sigma(self) -> bool:
        if (self.is_attribute("nu") & self.is_attribute("gamma") &
            self.is_attribute("gn_intensity_sigma")):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.gn_intensity_sigma
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_sigma"] = "\n".join(ls_out)


    def recalc_to_gamma_nu_grid(self):
        l_tth_grid = numpy.array(self.ttheta)*numpy.pi/180.
        l_phi_grid = numpy.array(self.phi)*numpy.pi/180.
        int_u = numpy.array(self.intensity_plus, dtype=float).transpose()
        int_d = numpy.array(self.intensity_minus, dtype=float).transpose()
        int_sum = int_u + int_d
        int_diff = int_u - int_d

        min_tth, max_tth = min(l_tth_grid), max(l_tth_grid)
        min_phi, max_phi = min(l_phi_grid), max(l_phi_grid)

        min_gamma, max_gamma = min_tth, max_tth
        num_gamma = len(l_tth_grid)

        min_nu, max_nu = -10.*numpy.pi/180., 15.*numpy.pi/180.
        num_nu = len(l_phi_grid)

        l_gamma_grid = numpy.linspace(min_gamma, max_gamma, num=num_gamma)
        l_nu_grid = numpy.linspace(min_nu, max_nu, num=num_nu)

        int_u_out = numpy.array(recal_int_to_gammanu_grid(l_tth_grid, l_phi_grid, int_u, l_gamma_grid, l_nu_grid), dtype=float)
        int_d_out = numpy.array(recal_int_to_gammanu_grid(l_tth_grid, l_phi_grid, int_d, l_gamma_grid, l_nu_grid), dtype=float)
        int_sum_out = numpy.array(recal_int_to_gammanu_grid(l_tth_grid, l_phi_grid, int_sum, l_gamma_grid, l_nu_grid), dtype=float)
        int_diff_out = numpy.array(recal_int_to_gammanu_grid(l_tth_grid, l_phi_grid, int_diff, l_gamma_grid, l_nu_grid), dtype=float)

        return l_gamma_grid*180./numpy.pi, l_nu_grid*180./numpy.pi, [int_u_out, int_d_out, int_sum_out, int_diff_out]
        
    def plots(self):
        return [self.plot_gamma_nu(), ]

    def plot_gamma_nu(self):
        flag_polarized = (self.is_attribute("gamma") & self.is_attribute("nu") &
               self.is_attribute("gn_intensity_plus") & 
               self.is_attribute("gn_intensity_minus"))
        flag_unpolarized = (self.is_attribute("gamma") & self.is_attribute("nu") &
               self.is_attribute("gn_intensity"))

        if not(self.is_attribute("gamma") & self.is_attribute("nu")):
            return

        gamma = self.gamma 
        nu = self.nu 
        if flag_polarized:
            signal_exp_plus = self.gn_intensity_plus
            signal_exp_minus = self.gn_intensity_minus
            signal_exp_plus_sigma = self.gn_intensity_plus_sigma
            signal_exp_minus_sigma = self.gn_intensity_minus_sigma
            signal_exp_sum = signal_exp_plus + signal_exp_minus
            signal_exp_difference = signal_exp_plus - signal_exp_minus
            signal_exp_sum_sigma_sq = numpy.square(signal_exp_plus_sigma)+numpy.square(signal_exp_minus_sigma)
        elif flag_unpolarized:
            signal_exp_sum = self.gn_intensity
            signal_exp_sum_sigma_sq = numpy.square(self.gn_intensity_sigma)
        else:
            return

        max_val = numpy.nanmax(signal_exp_sum)
        min_val = numpy.nanmin(signal_exp_sum)
        max_val = max_val - (max_val-min_val)*0.25

        gamma_rad, nu_rad = gamma*numpy.pi/180., nu*numpy.pi/180.
        ttheta_rad, phi_rad = calc_ttheta_phi_by_gamma_nu(
            gamma_rad[:, na], nu_rad[na, :], flag_gamma=False, flag_nu=False)[:2]
        ttheta_2d = ttheta_rad * 180./numpy.pi

        ttheta_step = numpy.round(gamma[1]-gamma[0], decimals=5)
        ttheta_min = gamma.min()

        ttheta_index = numpy.round((ttheta_2d-ttheta_min)/ttheta_step, decimals=0).astype(int)

        ttheta = numpy.linspace(
            ttheta_min,
            ttheta_step*ttheta_index.max()+ttheta_min,
            ttheta_index.max(), endpoint=False)

        signal_projection_exp_sum = numpy.zeros_like(ttheta)
        signal_projection_exp_difference = numpy.zeros_like(ttheta)
        signal_projection_exp_sum_sigma = numpy.zeros_like(ttheta)
        signal_projection_exp_difference_sigma = numpy.zeros_like(ttheta)

        if flag_polarized:
            flag_nonan = numpy.logical_not(numpy.isnan(signal_exp_difference))
        elif flag_unpolarized:
            flag_nonan = numpy.logical_not(numpy.isnan(signal_exp_sum))
        for ind in range(ttheta.size):
            flag_1 = ttheta_index == ind
            flag = numpy.logical_and(flag_1, flag_nonan)
            points = numpy.sum(flag)
            if points != 0:
                signal_projection_exp_sum[ind] += numpy.sum(signal_exp_sum[flag])/points
                signal_projection_exp_sum_sigma[ind] += numpy.sqrt(numpy.sum(signal_exp_sum_sigma_sq[flag]))/points
            else:
                signal_projection_exp_sum[ind] = numpy.nan
                signal_projection_exp_sum_sigma[ind] = numpy.nan

            if flag_polarized:
                flag = numpy.logical_and(flag_1, flag_nonan)
                points = numpy.sum(flag)
                if points != 0:
                    signal_projection_exp_difference[ind] += numpy.sum(signal_exp_difference[flag])/points
                    signal_projection_exp_difference_sigma[ind] += numpy.sqrt(numpy.sum(signal_exp_sum_sigma_sq[flag]))/points
                else:
                    signal_projection_exp_difference[ind] = numpy.nan
                    signal_projection_exp_difference_sigma[ind] = numpy.nan


        extent = [gamma.min(), gamma.max(), nu.min(), nu.max()]

        cmap_sum = plt.get_cmap("turbo") # BuPu
        if flag_polarized:
            fig, axs = plt.subplots(2, 2, sharex=True)
            ax1, ax2, ax3, ax4 = axs[0,0], axs[0,1], axs[1,0], axs[1,1]
        elif flag_unpolarized:
            fig, axs = plt.subplots(2, 1, sharex=True)
            ax1, ax3= axs[0], axs[1]

        norm_1 = matplotlib.colors.Normalize(vmax=max_val, vmin=min_val)
        hh = numpy.copy(signal_exp_sum)
        hh[numpy.isnan(hh)]=0
        ax1.imshow(hh.transpose(), origin="lower", aspect="auto", cmap=cmap_sum, norm= norm_1, extent=extent)

        ax1.set_xticks([])
        ax1.set_xlabel(r"$\gamma$ (deg.)")
        ax1.set_yticks([])
        ax1.set_ylabel(r"$\nu$ (deg.)")

        if flag_polarized:
            cmap_difference = plt.get_cmap("turbo") # BrBG
            max_val = numpy.nanmax(numpy.abs(signal_exp_difference))*0.50
            norm_2 = matplotlib.colors.Normalize(vmax=max_val, vmin=-max_val)
            hh = numpy.copy(signal_exp_difference)
            hh[numpy.isnan(hh)]=0
            ax2.imshow(hh.transpose(), origin="lower", aspect="auto",
                      norm=norm_2,
                      cmap=cmap_difference,
                      extent=extent)
            # ax2.imshow(zz_difference, origin="lower", aspect="auto", alpha=0.8, extent=extent)
            # ax2.contour(zz_difference[:,:,3], levels=[50, 70])

            # ax2.get_xaxis().set_visible(False)
            # ax2.get_yaxis().set_visible(False)
            ax2.set_xticks([])
            ax2.set_xlabel(r"$\gamma$ (deg.)")
            ax2.set_yticks([])
            ax2.set_ylabel(r"$\nu$ (deg.)")

        ax1.set_title(r"Unpolarized signal")


        ax3.fill_between(
            ttheta[:gamma.size],(signal_projection_exp_sum-signal_projection_exp_sum_sigma)[:gamma.size],
            (signal_projection_exp_sum+signal_projection_exp_sum_sigma)[:gamma.size], color="k", alpha=0.4)

        # ax3.errorbar(
        #     ttheta[:gamma.size], signal_projection_exp_sum[:gamma.size],
        #     yerr=signal_projection_exp_sum_sigma[:gamma.size], fmt="ko", alpha=0.2)
        # ax3.legend()
        ax3.set_xlabel(r"$2\theta$ (deg.)")

        if flag_polarized:
            ax2.set_title(r"Polarized signal")

            ax4.fill_between(
                ttheta[:gamma.size],(signal_projection_exp_difference-signal_projection_exp_difference_sigma)[:gamma.size],
                (signal_projection_exp_difference+signal_projection_exp_difference_sigma)[:gamma.size], color="k", alpha=0.4)

            # ax4.errorbar(
            #     ttheta[:gamma.size], signal_projection_exp_difference[:gamma.size],
            #     yerr=signal_projection_exp_difference_sigma[:gamma.size], fmt="ko", alpha=0.2)

            # ax4.legend()
            ax4.set_xlabel(r"$2\theta$ (deg.)")
        
        return (fig, ax1)




from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary
from cryspy.A_functions_base.function_3_extinction import \
    calc_extinction_2

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Extinction(ItemN):
    """
    Extinction class.

    Attributes
    ----------
        - mosaicity 100.0 # in minutes
        - radius    50.0 # in micrometers
        - model     "gauss" or "lorentz"



    """

    ATTR_MANDATORY_NAMES = ("model", "mosaicity", "radius")
    ATTR_MANDATORY_TYPES = (str, float, float)
    ATTR_MANDATORY_CIF = ("model", "mosaicity", "radius")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("mosaicity", "radius")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {"model": ["gauss", "lorentz"]}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "extinction"

    def __init__(self, **kwargs) -> NoReturn:
        super(Extinction, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"mosaicity": 0., "radius": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_extinction(self, cell, h, k, l, f_sq, wavelength:float,
                        flag_derivatives:bool=False):
        """
        f_sq in 10-12cm
        extinction for spherical model

        Parameters
        ----------
        cell : TYPE
            DESCRIPTION.
        h : TYPE
            DESCRIPTION.
        k : TYPE
            DESCRIPTION.
        l : TYPE
            DESCRIPTION.
        f_sq : TYPE
            DESCRIPTION.
        wavelength : TYPE
            DESCRIPTION.
        flag_derivatives : TYPE, optional
            DESCRIPTION. The default is False.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        radius, mosaicity, model = self.radius, self.mosaicity, self.model
        volume_unit_cell = cell.volume
        sthovl = cell.calc_sthovl(h, k, l)
        y_ext, dder = calc_extinction_2(radius, mosaicity, model, f_sq,
                                        volume_unit_cell, sthovl, wavelength)
        if flag_derivatives:
            return y_ext, dder
        else:
            return y_ext

    def get_dictionary(self):
        res = {}
        model_extinction = self.model
        radius = self.radius
        mosaicity = self.mosaicity
        res["extinction_model"] = model_extinction
        res["extinction_radius"] = numpy.array([radius], dtype=float)
        res["extinction_mosaicity"] = numpy.array([mosaicity], dtype=float)
        res["flags_extinction_radius"] = numpy.array([self.radius_refinement], dtype=bool)
        res["flags_extinction_mosaicity"] = numpy.array([self.mosaicity_refinement], dtype=bool)
        return res

class ExtinctionL(LoopN):
    """
    Description of extinction in loop.

    """
    ITEM_CLASS = Extinction
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(ExtinctionL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

# s_cont = """
#   loop_
#  _extinction_mosaicity
#  _extinction_radius
#  _extinction_model
#  100() 50 gauss
#  3 7() lorentz
# """

# obj = ExtinctionL.from_cif(s_cont)
# print(obj, end="\n\n")
